# -*- coding: utf-8 -*-:
"""
Functional tests for all modules in step_modules subpackage.
"""

from procars.step_modules import compute_conserved_adjacencies
from procars.step_modules import compute_resolved_conflicts
from procars.step_modules import compute_pqtree
from procars.step_modules import compute_initialpqtree
from procars.step_modules import compute_dcj_adjacencies
import pytest
import os
import shutil
try:
    import cPickle as pickle
except ImportError:
    import pickle


# ########################################################################################
# All fixtures needed for tests
@pytest.fixture(scope="module")
def expect_tree():
    return {1: ['N1', -1, 2, 7], 2: ['N2', 1, 3, 4], 3: ['C_c', 2, -1, -1], 4: ['N4', 2, 5, 6],
            5: ['A_a', 4, -1, -1], 6: ['B', 4, -1, -1], 7: ['N7', 1, 8, 9],
            8: ['D_species', 7, -1, -1], 9: ['E', 7, -1, -1]}


@pytest.fixture(scope="module")
def expect_blocks():
    return [[[], [[1, 99, 1, 1], [101, 199, 2, 1], [201, 299, 3, 1], [301, 399, 4, 1]],
             [[1, 99, 5, 1], [101, 199, 6, 1], [201, 299, 7, 1], [301, 399, 8, -1]]],
            [[], [[1, 99, 1, 1], [101, 199, 2, 1], [201, 299, 3, 1], [301, 399, 4, 1],
             [401, 499, 5, 1], [501, 599, 6, 1], [601, 699, 7, 1], [701, 799, 8, 1]]],
            [[], [[1, 99, 1, 1], [101, 199, 2, 1], [201, 299, 3, 1], [301, 399, 4, 1],
             [401, 499, 5, 1], [501, 599, 6, 1], [601, 699, 7, 1], [701, 799, 8, -1]]],
            [[], [[1, 99, 1, 1], [101, 199, 2, -1], [201, 299, 3, 1], [301, 399, 5, -1]],
             [[1, 99, 8, 1], [101, 199, 7, -1], [201, 299, 6, -1], [301, 399, 4, 1]]],
            [[], [[1, 99, 1, 1], [101, 199, 2, 1], [201, 299, 3, 1], [301, 399, 5, -1]],
             [[1, 99, 4, -1], [101, 199, 6, 1], [201, 299, 7, 1], [301, 399, 8, 1]]]]


# ########################################################################################
# Functional tests
def test_compute_conserved():
    """
    Test main method of compute_conserved_adjacencies
    """
    DATA_DIR = os.path.join("test", "data")
    PQTREE_FILE = os.path.join(DATA_DIR, 'res_PQtree')
    ADJACENCY_FILE = os.path.join(DATA_DIR, 'res_func.conserved_adjacencies')
    tree_bin = os.path.join(DATA_DIR, "res_tree_information.bin")
    blocks_bin = os.path.join(DATA_DIR, "res_block_information.bin")
    car_bin = os.path.join(DATA_DIR, "res_car_information.bin")
    car_neigh_bin = os.path.join(DATA_DIR, "res_neigh_car_information.bin")
    step_nb = 1
    # create empty file of adjacencies for last step (step 0)
    res_last = os.path.join(DATA_DIR, 'res_func.conserved_adjacencies_0.txt')
    open(res_last, "w").close()
    (nb_Final_Adj, nb_discarded, FS_NC_Adj, PS_NC2_Adj, FS_C_Adj, PS_NC_C_Adj,
     PS_C_Adj) = compute_conserved_adjacencies.main(PQTREE_FILE, ADJACENCY_FILE, step_nb,
                                                    tree_bin, blocks_bin, car_bin, car_neigh_bin)
    assert nb_Final_Adj == 5
    assert nb_discarded == 2
    assert FS_NC_Adj == [[6, 7]]
    assert PS_NC2_Adj == [[1, 2], [2, 3], [3, -5], [-4, 6]]
    assert FS_C_Adj == []
    assert PS_NC_C_Adj == []
    assert PS_C_Adj == [[7, 8], [7, -8]]
    os.remove(res_last)
    res_new = os.path.join(DATA_DIR, 'res_func.conserved_adjacencies_1.txt')
    os.remove(res_new)
    os.remove(res_new + "_discarded.txt")


def test_resolve_conflicts():
    """
    Test on main method of compute_resolved_conflicts
    """
    step = 2
    reffile = os.path.join("test", "data",
                           "res_S_Adjacencies_boreoeutherian_ancestor_2.txt_discarded.txt")
    testfile = os.path.join("test", "data", "S_adj_test_main_resolveconflict_2.txt_discarded.txt")
    adjfile = os.path.join("test", "data", "S_adj_test_main_resolveconflict_1.txt")
    # create empty file for adjacencies added in previous steps
    open(adjfile, "w").close()
    shutil.copyfile(reffile, testfile)
    adj_file = 'test/data/S_adj_test_main_resolveconflict'
    tree_bin = os.path.join("test", "data", "res_tree_information.bin")
    max_set, ind_set = compute_resolved_conflicts.main(adj_file, step, tree_bin)
    assert max_set == [0, 3, 5]
    assert ind_set == [[0, 3, 5], [0, 4, 5], [1, 2, 3], [1, 2, 4]]
    os.remove(testfile)
    sadjfile = os.path.join("test", "data", "S_adj_test_main_resolveconflict_2.txt")
    os.remove(sadjfile)
    os.remove(adjfile)
    sadj_discard_used = os.path.join("test", "data",
                                     "S_adj_test_main_resolveconflict_2.txt_discarded_used.txt")
    os.remove(sadj_discard_used)


def test_compute_pqtree():
    """
    Test main method of compute_pqtree
    """
    nb_blocks = 689
    adjacency_prefix = os.path.join("test", "data", "res_test_computepqtree_adj")
    car_prefix = os.path.join("test", "data", "res_test_computepqtree_PQtree")
    step_nb = 2
    refpqtree = os.path.join("test", "data", "res_reftest_computepqtree_PQtree_2.txt")
    compute_pqtree.main(nb_blocks, adjacency_prefix, car_prefix, step_nb)
    with open(car_prefix + "_" + str(step_nb) + ".txt", "r") as result:
        all_lines = result.readlines()
    with open(refpqtree, "r") as reference:
        all_lines_ref = reference.readlines()
    assert all_lines == all_lines_ref
    os.remove(car_prefix + '_' + str(step_nb) + ".txt")


def test_compute_initialpqtree(expect_tree, expect_blocks):
    """
    Test main method of compute_initialpqtree
    """
    tree_file_name = os.path.join("test", "data", "data_tree.ph")
    block_file_name = os.path.join("test", "data", "data_orthology_blocks.txt")
    car_file_name = os.path.join("test", "data", "res_main_computeinitialpqtree_PQtree.txt")
    refpqtree0 = os.path.join("test", "data", "res_PQtree_0.txt")
    adj_file = os.path.join("test", "data", "res_main_computeinitialpqtree_adjs.txt")
    tree_bin = os.path.join("test", "data", "res_tree_information_main_computeinitialpqtree.bin")
    blocks_bin = os.path.join("test", "data", "res_block_information_main_computeinitialpqtree.bin")
    nb_blocks = compute_initialpqtree.main(tree_file_name, block_file_name, car_file_name,
                                           tree_bin, blocks_bin, adj_file)
    assert nb_blocks == 8
    with open(car_file_name, "r") as res:
        reslines = res.readlines()
    with open(refpqtree0, "r") as ref:
        reflines = ref.readlines()
    assert reslines == reflines
    with open(adj_file, "r") as adj:
        assert adj.readlines() == []
    with open(tree_bin, "rb") as treefile:
        my_dep = pickle.Unpickler(treefile)
        info = my_dep.load()
    assert info[0] == expect_tree
    with open(blocks_bin, "rb") as treefile:
        my_dep = pickle.Unpickler(treefile)
        infoblock = my_dep.load()
    assert infoblock[0] == expect_blocks
    os.remove(tree_bin)
    os.remove(blocks_bin)
    os.remove(car_file_name)
    os.remove(adj_file)


def test_computedcj():
    """
    Test on the main method of compute_dcj_adjacencies
    """
    tree_bin = os.path.join('test', 'data', "res_tree_boreo_information.bin")
    blocks_bin = os.path.join('test', 'data', 'res_block_boreo_information.bin')
    neigh_bin = os.path.join('test', 'data', 'res_neigh_dcj.bin')
    car_bin = os.path.join('test', 'data', 'res_cars_dcj.bin')
    adj_file = os.path.join('test', 'data', 'res_test_computedcj_adj')
    # create empty adjacency file for last adjacencies:
    open(adj_file + "_3.txt", "w").close()
    step_nb = 4
    dcj_NC_Adj, step_nb, dcj_NC_Pair = compute_dcj_adjacencies.main(adj_file, step_nb, car_bin,
                                                                    neigh_bin, tree_bin, blocks_bin)
    assert dcj_NC_Adj == [[2, -20], [23, 24], [29, 28]]
    assert step_nb == 4
    assert dcj_NC_Pair == [(2, 20), (23, 54), (29, 58)]
    os.remove(adj_file + "_4.txt")
    os.remove(adj_file + "_3.txt")
