# -*- coding: utf-8 -*-:
"""
Unit tests for compute_conserved_adjacencies.py
"""
from procars.step_modules import compute_conserved_adjacencies
from procars.utils.util_classes import SparseMatrix
import pytest
import StringIO
import os
try:
    import cPickle as pickle
except ImportError:
    import pickle


# ########################################################################################
# All fixtures needed for tests
@pytest.fixture(scope="module")
def treebin():
    return os.path.join("test", "data", "res_tree_information.bin")


@pytest.fixture(scope="module")
def blocksbin():
    return os.path.join("test", "data", "res_block_information.bin")


@pytest.fixture(scope="module")
def expect_tree():
    return {1: ['N1', -1, 2, 7], 2: ['N2', 1, 3, 4], 3: ['C_c', 2, -1, -1], 4: ['N4', 2, 5, 6],
            5: ['A_a', 4, -1, -1], 6: ['B', 4, -1, -1], 7: ['N7', 1, 8, 9],
            8: ['D_species', 7, -1, -1], 9: ['E', 7, -1, -1]}


@pytest.fixture(scope="module")
def expect_blocks():
    return [[[], [[1, 99, 1, 1], [101, 199, 2, 1], [201, 299, 3, 1], [301, 399, 4, 1]],
             [[1, 99, 5, 1], [101, 199, 6, 1], [201, 299, 7, 1], [301, 399, 8, -1]]],
            [[], [[1, 99, 1, 1], [101, 199, 2, 1], [201, 299, 3, 1], [301, 399, 4, 1],
             [401, 499, 5, 1], [501, 599, 6, 1], [601, 699, 7, 1], [701, 799, 8, 1]]],
            [[], [[1, 99, 1, 1], [101, 199, 2, 1], [201, 299, 3, 1], [301, 399, 4, 1],
             [401, 499, 5, 1], [501, 599, 6, 1], [601, 699, 7, 1], [701, 799, 8, -1]]],
            [[], [[1, 99, 1, 1], [101, 199, 2, -1], [201, 299, 3, 1], [301, 399, 5, -1]],
             [[1, 99, 8, 1], [101, 199, 7, -1], [201, 299, 6, -1], [301, 399, 4, 1]]],
            [[], [[1, 99, 1, 1], [101, 199, 2, 1], [201, 299, 3, 1], [301, 399, 5, -1]],
             [[1, 99, 4, -1], [101, 199, 6, 1], [201, 299, 7, 1], [301, 399, 8, 1]]]]


# ########################################################################################
# Unit tests on compute_conserved_adjacencies
def test_compute_blocks_cars(expect_blocks):
    """
    Test on finding the blocks' cars of each species
    """
    block_2_car = [0, 1, 2, 3, 4, 5, 6, 7, 8]
    nb_species = 5
    blocks_car = compute_conserved_adjacencies.compute_blocks_cars(nb_species,
                                                                   expect_blocks, block_2_car)
    assert blocks_car == [[[], [1, 2, 3, 4], [5, 6, 7, 8]], [[], [1, 2, 3, 4, 5, 6, 7, 8]],
                          [[], [1, 2, 3, 4, 5, 6, 7, 8]], [[], [1, 2, 3, 5], [8, 7, 6, 4]],
                          [[], [1, 2, 3, 5], [4, 6, 7, 8]]]


def test_commons_read_cars(expect_blocks):
    """
    Test on reading and finding all information on Cars
    """
    nb_blocks = 8
    nb_species = 5
    filename = os.path.join("test", "data", "res_PQtree")
    step_nb = 1
    binfile = os.path.join("test", "data", "car_information.bin")
    ref_binfile = os.path.join("test", "data", "res_car_information.bin")
    res = compute_conserved_adjacencies.commons_read_cars(nb_blocks, expect_blocks, nb_species,
                                                          filename, step_nb, binfile)
    cars, nb_cars, block_position_in_car, blocks_car = res
    with open(binfile, "rb") as blockfile:
        my_dep = pickle.Unpickler(blockfile)
        info = my_dep.load()
    with open(ref_binfile, "rb") as refblockfile:
        my_dep = pickle.Unpickler(refblockfile)
        info_ref = my_dep.load()
    for index, inf in enumerate(info):
        assert inf == info_ref[index]
    os.remove(binfile)
    assert cars == [[], [1], [2], [3], [4], [5], [6], [7], [8]]
    assert nb_cars == 8
    assert block_position_in_car == [0] * 9
    assert blocks_car == [[[], [1, 2, 3, 4], [5, 6, 7, 8]], [[], [1, 2, 3, 4, 5, 6, 7, 8]],
                          [[], [1, 2, 3, 4, 5, 6, 7, 8]], [[], [1, 2, 3, 5], [8, 7, 6, 4]],
                          [[], [1, 2, 3, 5], [4, 6, 7, 8]]]


def test_potential_neighbor(expect_blocks):
    """
    Test on finding potential car neighbors
    """
    spe_id = 0
    bl_car = [[[], [1, 2, 3, 4], [5, 6, 7, 8]], [[], [1, 2, 3, 4, 5, 6, 7, 8]],
              [[], [1, 2, 3, 4, 5, 6, 7, 8]], [[], [1, 2, 3, 5], [8, 7, 6, 4]],
              [[], [1, 2, 3, 5], [4, 6, 7, 8]]]
    res = compute_conserved_adjacencies.compute_potential_cars_neighbors(spe_id, expect_blocks,
                                                                         bl_car)
    neighbors, segments_neighbors = res
    assert neighbors == [[], [1, 2], [2, 3], [3, 4], [5, 6], [6, 7], [7, 8]]
    assert segments_neighbors == [[], [[1], [2]], [[2], [3]], [[3], [4]], [[5], [6]],
                                  [[6], [7]], [[7], [-8]]]


def test_check_potential_cars_neighbors():
    """
    Test on checking potential neighboring cars
    """
    spe_id = 0
    neighbors = [[], [1, 2], [2, 3], [3, 4], [5, 6], [6, 7], [7, 8]]
    segments_neighbors = [[], [[1], [2]], [[2], [3]], [[3], [4]], [[5], [6]],
                          [[6], [7]], [[7], [-8]]]
    cars = [[], [1], [2], [3], [4], [5], [6], [7], [8]]
    block_position_in_car = [0] * 9
    left = [[[], [], [1], [2], [3], [], [5], [6], []],
            [[], [], [1], [2], [3], [4], [5], [6], [7]],
            [[], [], [1], [2], [3], [4], [5], [6], []],
            [[], [], [-3], [-2], [-6], [], [-4], [6], []],
            [[], [], [1], [2], [-6], [], [-4], [6], [7]]]
    right = [[[], [2], [3], [4], [], [6], [7], [-8], [-7]],
             [[], [2], [3], [4], [5], [6], [7], [8], []],
             [[], [2], [3], [4], [5], [6], [7], [-8], [-7]],
             [[], [-2], [-1], [-5], [], [-3], [7], [-8], [-7]],
             [[], [2], [3], [-5], [], [-3], [7], [8], []]]
    res = compute_conserved_adjacencies.check_potential_cars_neighbors(spe_id, neighbors,
                                                                       segments_neighbors, cars,
                                                                       block_position_in_car,
                                                                       left, right)
    left, right = res
    assert left == [[[], [], [1, 1], [2, 2], [3, 3], [], [5, 5], [6, 6], []],
                    [[], [], [1], [2], [3], [4], [5], [6], [7]],
                    [[], [], [1], [2], [3], [4], [5], [6], []],
                    [[], [], [-3], [-2], [-6], [], [-4], [6], []],
                    [[], [], [1], [2], [-6], [], [-4], [6], [7]]]
    assert right == [[[], [2, 2], [3, 3], [4, 4], [], [6, 6], [7, 7], [-8, -8], [-7, -7]],
                     [[], [2], [3], [4], [5], [6], [7], [8], []],
                     [[], [2], [3], [4], [5], [6], [7], [-8], [-7]],
                     [[], [-2], [-1], [-5], [], [-3], [7], [-8], [-7]],
                     [[], [2], [3], [-5], [], [-3], [7], [8], []]]


def test_compute_carneighs(expect_blocks):
    """
    Test on finding all possible neighbors for each car
    """
    # species tree = ((C,(A,B)),(D,E)@);
    # spe_ids = A:1, B:2, C:0, D: 3, E: 4
    # genomes : A= [1, 2, 3, 4, 5, 6, 7, 8]
    #           B= [1, 2, 3, 4, 5, 6, 7, -8]
    #           C= [1, 2, 3, 4], [5, 6, 7, -8]
    #           D= [1, -2, 3, -5], [8, -7, 6, 4]
    #           E= [1, 2, 3, -5], [-4, 6, 7, 8]
    nb_cars = 6
    nb_species = 5
    blocks = expect_blocks
    cars = [[], [1, 2, 3], [4], [5], [6], [7], [8]]
    bl_car = [[[], [1, 1, 1, 3], [3, 4, 5, 6]], [[], [1, 1, 1, 2, 3, 4, 5, 6]],
              [[], [1, 1, 1, 2, 3, 4, 5, 6]], [[], [1, 1, 1, 3], [6, 5, 4, 2]],
              [[], [1, 1, 1, 3], [2, 4, 5, 6]]]
    block_position_in_car = [0, 0, 1, 2, 0, 0, 0, 0, 0]
    binary_file_name = os.path.join("test", "data", "plop.bin")
    le, ri = compute_conserved_adjacencies.compute_cars_neighbors(nb_cars, nb_species, blocks,
                                                                  cars, bl_car,
                                                                  block_position_in_car,
                                                                  binary_file_name)
    assert le == [[[], [], [], [], [3], [4], []], [[], [], [1], [2], [3], [4], [5]],
                  [[], [], [1], [2], [3], [4], []], [[], [], [-4], [], [-2], [4], []],
                  [[], [], [-4], [], [-2], [4], [5]]]
    assert ri == [[[], [], [], [4], [5], [-6], [-5]], [[], [2], [3], [4], [5], [6], []],
                  [[], [2], [3], [4], [5], [-6], [-5]], [[], [-3], [], [-1], [5], [-6], [-5]],
                  [[], [-3], [], [-1], [5], [6], []]]
    with open(binary_file_name, "rb") as carfile:
        my_dep = pickle.Unpickler(carfile)
        info = my_dep.load()
    assert info[0] == le
    assert info[1] == ri
    os.remove(binary_file_name)


def test_fill_conservedarrays_fully():
    """
    test about filling conservation status arrays
    for an adjacency fully conserved (1 2)
    """
    # species tree = ((C,(A,B)),(D,E)@);
    # spe_ids = A:1, B:2, C:0, D: 3, E: 4
    # genomes : A= [1, 2, 3, 4, 5, 6, 7, 8]
    #           B= [1, 2, 3, 4, 5, 6, 7, -8]
    #           C= [1, 2, 3, 4], [5, 6, 7, -8]
    #           D= [1, 2, 3, -5], [8, -7, 6, 4]
    #           E= [1, 2, 3, -5], [-4, 6, 7, 8]
    nb_blocks = 8
    conserved_left_D = [[] for _ in xrange(nb_blocks + 1)]
    conserved_right_D = [[] for _ in xrange(nb_blocks + 1)]
    conserved_left = [[] for _ in xrange(nb_blocks + 1)]
    conserved_right = [[] for _ in xrange(nb_blocks + 1)]
    left = [[[], [], [1], [2], [3], [], [5], [6], []],
            [[], [], [1], [2], [3], [4], [5], [6], [7]],
            [[], [], [1], [2], [3], [4], [5], [6], []],
            [[], [], [1], [2], [6], [], [-7], [-6], []],
            [[], [], [1], [2], [-6], [], [-4], [6], [7]]]
    right = [[[], [2], [3], [4], [], [6], [7], [-8], [-7]],
             [[], [2], [3], [4], [5], [6], [7], [8], []],
             [[], [2], [3], [4], [5], [6], [7], [-8], [-7]],
             [[], [2], [3], [-5], [], [-3], [4], [-8], []],
             [[], [2], [3], [-5], [], [-3], [7], [8], []]]
    species1_id = [0, 1, 2]
    species2_id = [3]
    species3_id = [4]
    block1 = 1
    block2 = 2
    compute_conserved_adjacencies.fill_conserved_arrays(conserved_left_D, conserved_right_D,
                                                        conserved_left, conserved_right,
                                                        left, right, species1_id, species2_id,
                                                        species3_id, block1, block2)
    assert conserved_left_D == [[], [], [1], [], [], [], [], [], []]
    assert conserved_right_D == [[], [2], [], [], [], [], [], [], []]
    assert conserved_left == [[] for _ in xrange(nb_blocks + 1)]
    assert conserved_right == [[] for _ in xrange(nb_blocks + 1)]


def test_fill_conservedarrays_fully2():
    """
    test about filling conservation status arrays
    for an adjacency fully conserved (-2 -1)
    """
    # species tree = ((C,(A,B)),(D,E)@);
    # spe_ids = A:1, B:2, C:0, D: 3, E: 4
    # genomes : A= [1, 2, 3, 4, 5, 6, 7, 8]
    #           B= [1, 2, 3, 4, 5, 6, 7, -8]
    #           C= [1, 2, 3, 4], [5, 6, 7, -8]
    #           D= [1, 2, 3, -5], [8, -7, 6, 4]
    #           E= [1, 2, 3, -5], [-4, 6, 7, 8]
    nb_blocks = 8
    conserved_left_D = [[] for _ in xrange(nb_blocks + 1)]
    conserved_right_D = [[] for _ in xrange(nb_blocks + 1)]
    conserved_left = [[] for _ in xrange(nb_blocks + 1)]
    conserved_right = [[] for _ in xrange(nb_blocks + 1)]
    left = [[[], [], [1], [2], [3], [], [5], [6], []],
            [[], [], [1], [2], [3], [4], [5], [6], [7]],
            [[], [], [1], [2], [3], [4], [5], [6], []],
            [[], [], [1], [2], [6], [], [-7], [-6], []],
            [[], [], [1], [2], [-6], [], [-4], [6], [7]]]
    right = [[[], [2], [3], [4], [], [6], [7], [-8], [-7]],
             [[], [2], [3], [4], [5], [6], [7], [8], []],
             [[], [2], [3], [4], [5], [6], [7], [-8], [-7]],
             [[], [2], [3], [-5], [], [-3], [4], [-8], []],
             [[], [2], [3], [-5], [], [-3], [7], [8], []]]
    species1_id = [0, 1, 2]
    species2_id = [3]
    species3_id = [4]
    block1 = -2
    block2 = -1
    compute_conserved_adjacencies.fill_conserved_arrays(conserved_left_D, conserved_right_D,
                                                        conserved_left, conserved_right,
                                                        left, right, species1_id, species2_id,
                                                        species3_id, block1, block2)
    assert conserved_left_D == [[], [], [1], [], [], [], [], [], []]
    assert conserved_right_D == [[], [2], [], [], [], [], [], [], []]
    assert conserved_left == [[] for _ in xrange(nb_blocks + 1)]
    assert conserved_right == [[] for _ in xrange(nb_blocks + 1)]


def test_fill_conservedarrays_partly():
    """
    test about filling conservation status arrays
    for an adjacency partly conserved (-8 -7)
    """
    # same data set as in test_fill_conservedarrays
    nb_blocks = 8
    conserved_left_D = [[] for _ in xrange(nb_blocks + 1)]
    conserved_right_D = [[] for _ in xrange(nb_blocks + 1)]
    conserved_left = [[] for _ in xrange(nb_blocks + 1)]
    conserved_right = [[] for _ in xrange(nb_blocks + 1)]
    left = [[[], [], [1], [2], [3], [], [5], [6], []],
            [[], [], [1], [2], [3], [4], [5], [6], [7]],
            [[], [], [1], [2], [3], [4], [5], [6], []],
            [[], [], [1], [2], [6], [], [-7], [-6], []],
            [[], [], [1], [2], [-6], [], [-4], [6], [7]]]
    right = [[[], [2], [3], [4], [], [6], [7], [-8], [-7]],
             [[], [2], [3], [4], [5], [6], [7], [8], []],
             [[], [2], [3], [4], [5], [6], [7], [-8], [-7]],
             [[], [2], [3], [-5], [], [-3], [4], [-8], []],
             [[], [2], [3], [-5], [], [-3], [7], [8], []]]
    species1_id = [0, 1, 2]
    species2_id = [3]
    species3_id = [4]
    block1 = -8
    block2 = -7
    compute_conserved_adjacencies.fill_conserved_arrays(conserved_left_D, conserved_right_D,
                                                        conserved_left, conserved_right,
                                                        left, right, species1_id, species2_id,
                                                        species3_id, block1, block2)
    assert conserved_left == [[], [], [], [], [], [], [], [], [7]]
    assert conserved_right == [[], [], [], [], [], [], [], [8], []]
    assert conserved_left_D == [[] for _ in xrange(nb_blocks + 1)]
    assert conserved_right_D == [[] for _ in xrange(nb_blocks + 1)]


def test_fill_conservedarrays_partly2():
    """
    test about filling conservation status arrays
    for an adjacency partly conserved (7 8)
    """
    # same data set as in test_fill_conservedarrays
    nb_blocks = 8
    conserved_left_D = [[] for _ in xrange(nb_blocks + 1)]
    conserved_right_D = [[] for _ in xrange(nb_blocks + 1)]
    conserved_left = [[] for _ in xrange(nb_blocks + 1)]
    conserved_right = [[] for _ in xrange(nb_blocks + 1)]
    left = [[[], [], [1], [2], [3], [], [5], [6], []],
            [[], [], [1], [2], [3], [4], [5], [6], [7]],
            [[], [], [1], [2], [3], [4], [5], [6], []],
            [[], [], [1], [2], [6], [], [-7], [-6], []],
            [[], [], [1], [2], [-6], [], [-4], [6], [7]]]
    right = [[[], [2], [3], [4], [], [6], [7], [-8], [-7]],
             [[], [2], [3], [4], [5], [6], [7], [8], []],
             [[], [2], [3], [4], [5], [6], [7], [-8], [-7]],
             [[], [2], [3], [-5], [], [-3], [4], [-8], []],
             [[], [2], [3], [-5], [], [-3], [7], [8], []]]
    species1_id = [0, 1, 2]
    species2_id = [3]
    species3_id = [4]
    block1 = 7
    block2 = 8
    compute_conserved_adjacencies.fill_conserved_arrays(conserved_left_D, conserved_right_D,
                                                        conserved_left, conserved_right,
                                                        left, right, species1_id, species2_id,
                                                        species3_id, block1, block2)
    assert conserved_left == [[], [], [], [], [], [], [], [], [7]]
    assert conserved_right == [[], [], [], [], [], [], [], [8, -8], [-7]]
    assert conserved_left_D == [[] for _ in xrange(nb_blocks + 1)]
    assert conserved_right_D == [[] for _ in xrange(nb_blocks + 1)]


def test_reverse_seg():
    """
    Test on reversing a segment of CARs
    """
    segment = [1, -2, -3, 4, 8, 9, 10, 5, -13]
    reversed_seg = compute_conserved_adjacencies.reverse(segment)
    assert reversed_seg == [13, -5, -10, -9, -8, -4, 3, 2, -1]


def test_share_extr():
    """
    Tests if a segment and a car share a same extremity region in terms of synteny.
    """
    segment = [1, -2, 4, -3]
    car = [1, 2, 3, 4]
    block_pos = [0, 0, 1, 2, 3]
    side_segment = 1
    side_car = 1
    res = compute_conserved_adjacencies.segment_car_common_extremity(segment, car, block_pos,
                                                                     side_segment, side_car)
    assert res


def test_share_extr2():
    """
    Tests if a segment and a car share a same extremity region in terms of synteny.
    """
    segment = [1, -2, 4, -3]
    car = [-3, 4, 2, -1]
    block_pos = [0, 3, 2, 0, 1]
    side_segment = 1
    side_car = 0
    res = compute_conserved_adjacencies.segment_car_common_extremity(segment, car, block_pos,
                                                                     side_segment, side_car)
    assert res


def test_notshare_extr():
    """
    Tests if a segment and a car share a same extremity region in terms of synteny.
    """
    segment = [3, 1, -4, 2]
    car = [1, 2, 3, 4]
    block_pos = [0, 0, 1, 2, 3]
    side_segment = 1
    side_car = 1
    res = compute_conserved_adjacencies.segment_car_common_extremity(segment, car, block_pos,
                                                                     side_segment, side_car)
    assert not res


def test_compute_extr_i():
    """
    Test on computing orientation of a car in a car adj
    last block of segment is the same as last block of given car -> extremity = given car
    """
    segment = [-2, 4]
    cars = [[1, -3], [2, 4]]
    car_id = 1
    block_position_in_car = [0, 0, 0, 1, 1]
    orientation = 1
    ext = compute_conserved_adjacencies.compute_extremity(segment, car_id, cars,
                                                          block_position_in_car, orientation)
    assert ext == 1


def test_compute_extr_i2():
    """
    Test on computing orientation of a car in a car adj
    last block of segment is the same as -first block of given car -> extremity = -given car
    """
    segment = [-2, 4]
    cars = [[2, -3], [-4, 1]]
    car_id = 1
    block_position_in_car = [0, 0, 0, 1, 1]
    orientation = 1
    ext = compute_conserved_adjacencies.compute_extremity(segment, car_id, cars,
                                                          block_position_in_car, orientation)
    assert ext == -1


def test_compute_extr_ii():
    """
    Test on computing orientation of a car in a car adj
    abs(last block of segment) is the same as abs(last block of given car) -> extremity = given car
    """
    segment = [-2, -4]
    cars = [[2, -3], [1, 4]]
    car_id = 1
    block_position_in_car = [0, 0, 0, 1, 1]
    orientation = 1
    ext = compute_conserved_adjacencies.compute_extremity(segment, car_id, cars,
                                                          block_position_in_car, orientation)
    assert ext == 1


def test_compute_extr_ii2():
    """
    Test on computing orientation of a car in a car adj
    abs(last block of segment) is the same as abs(first block of given car)
    -> extremity = -given car
    """
    segment = [-2, -4]
    cars = [[2, -3], [-4, 1]]
    car_id = 1
    block_position_in_car = [0, 0, 0, 1, 1]
    orientation = 1
    ext = compute_conserved_adjacencies.compute_extremity(segment, car_id, cars,
                                                          block_position_in_car, orientation)
    assert ext == -1


def test_compute_extr_iii():
    """
    Test on computing orientation of a car in a car adj
    segment is syntenic to a segment of given car containing abs(last block of car)
    """
    segment = [8, -7, 10]
    cars = [[2, -3, 6, 5], [-4, 1, 9, -10, 8, 7]]
    car_id = 1
    block_position_in_car = [0, 1, 0, 1, 0, 3, 2, 5, 4, 2, 3]
    orientation = 1
    ext = compute_conserved_adjacencies.compute_extremity(segment, car_id, cars,
                                                          block_position_in_car, orientation)
    assert ext == 1


def test_compute_extr_iii2():
    """
    Test on computing orientation of a car in a car adj
    segment is syntenic to a segment of given car containing abs(first block of car)
    """
    segment = [-1, 4, 9]
    cars = [[2, -3, 6, 5], [-4, 1, 9, -10, 8, 7]]
    car_id = 1
    block_position_in_car = [0, 1, 0, 1, 0, 3, 2, 5, 4, 2, 3]
    orientation = -1
    ext = compute_conserved_adjacencies.compute_extremity(segment, car_id, cars,
                                                          block_position_in_car, orientation)
    assert ext == 1


def test_compute_conserved():
    """
    test on computing the status of all adjacencies
    """
    # species tree = ((C,(A,B)),(D,E)@);
    # spe_ids = A:1, B:2, C:0, D: 3, E: 4
    # genomes : A= [1, 2, 3, 4, 5, 6, 7, 8]
    #           B= [1, 2, 3, 4, 5, 6, 7, -8]
    #           C= [1, 2, 3, 4], [5, 6, 7, -8]
    #           D= [1, 2, 3, -5], [8, -7, 6, 4]
    #           E= [1, 2, 3, -5], [-4, 6, 7, 8]
    left = [[[], [], [1], [2], [3], [], [5], [6], []],
            [[], [], [1], [2], [3], [4], [5], [6], [7]],
            [[], [], [1], [2], [3], [4], [5], [6], []],
            [[], [], [1], [2], [6], [], [-7], [-6], []],
            [[], [], [1], [2], [-6], [], [-4], [6], [7]]]
    right = [[[], [2], [3], [4], [], [6], [7], [-8], [-7]],
             [[], [2], [3], [4], [5], [6], [7], [8], []],
             [[], [2], [3], [4], [5], [6], [7], [-8], [-7]],
             [[], [2], [3], [-5], [], [-3], [4], [-8], []],
             [[], [2], [3], [-5], [], [-3], [7], [8], []]]
    species1_id = [0, 1, 2]
    species2_id = [3]
    species3_id = [4]
    nb_blocks = 8
    res = compute_conserved_adjacencies.compute_conserved_adjacencies(left, right, nb_blocks,
                                                                      species1_id, species2_id,
                                                                      species3_id)
    conserved_left_D, conserved_right_D, conserved_left, conserved_right = res
    # fully conserved: (1 2) and (2 3)
    assert conserved_left_D == [[], [], [1], [2], [], [], [], [], []]
    assert conserved_right_D == [[], [2], [3], [], [], [], [], [], []]
    # partly conserved: (6 7) (7 8) (3 -5) (7 -8)
    assert conserved_left == [[], [], [], [], [], [], [], [6], [7]]
    assert conserved_right == [[], [], [], [-5], [], [-3], [7], [8, -8], [-7]]


def test_adj_matrix():
    """
    Test on creating the adjacency matrix
    """
    left = [[], [], [1], [2], [3], [], [5], [6], []]
    right = [[], [2], [3], [4], [], [6], [7], [-8], [-7]]
    nb_blocks = 8
    adj_mat = compute_conserved_adjacencies.compute_adjacencymatrix_from_neighbor_arrays(left,
                                                                                         right,
                                                                                         nb_blocks)
    mat = SparseMatrix(nb_blocks * 2 + 1, nb_blocks * 2 + 1)
    mat[1, 10] = 1
    mat[2, 11] = 1
    mat[3, 12] = 1
    mat[5, 14] = 1
    mat[6, 15] = 1
    mat[7, 8] = 1
    mat[8, 7] = 1
    mat[10, 1] = 1
    mat[11, 2] = 1
    mat[12, 3] = 1
    mat[14, 5] = 1
    mat[15, 6] = 1
    assert adj_mat == mat


def test_diff_ncmatrix():
    """
    Separate content of a matrix into two matrices : adjacencies in conflict with another matrix,
    and adjacencies compatible.
    """
    # fully conserved non-conflicting adjacencies matrix
    nb_blocks = 8
    mat = SparseMatrix(nb_blocks * 2 + 1, nb_blocks * 2 + 1)
    mat[1, 10] = 1
    mat[2, 11] = 1
    mat[3, 12] = 1
    mat[6, 15] = 1
    mat[7, 8] = 1
    mat[8, 7] = 1
    mat[10, 1] = 1
    mat[11, 2] = 1
    mat[12, 3] = 1
    mat[15, 6] = 1
    # partly conserved non-conflicting adjacencies matrix
    psmat = SparseMatrix(nb_blocks * 2 + 1, nb_blocks * 2 + 1)
    psmat[1, 9] = 1
    psmat[2, 10] = 1
    psmat[10, 2] = 1
    psmat[9, 1] = 1
    psmat[4, 5] = 1
    psmat[5, 4] = 1
    ncmat, cmat = compute_conserved_adjacencies.diff_nc_matrix_m(mat, psmat, nb_blocks)
    ex_ncmat = SparseMatrix(nb_blocks * 2 + 1, nb_blocks * 2 + 1)
    ex_ncmat[4, 5] = 1
    ex_ncmat[5, 4] = 1
    ex_cmat = SparseMatrix(nb_blocks * 2 + 1, nb_blocks * 2 + 1)
    ex_cmat[1, 9] = 1
    ex_cmat[2, 10] = 1
    ex_cmat[10, 2] = 1
    ex_cmat[9, 1] = 1
    assert ncmat == ex_ncmat
    assert cmat == ex_cmat


def test_findinfoandwriteadj():
    """
    test on writing adjs found into output file
    """
    # species tree = ((C,(A,B)),(D,E)@);
    # spe_ids = A:1, B:2, C:0, D: 3, E: 4
    # genomes : A= [1, 2, 3, 4, 5, 6, 7, 8]
    #           B= [1, 2, 3, 4, 5, 6, 7, -8]
    #           C= [1, 2, 3, 4], [5, 6, 7, -8]
    #           D= [1, 2, 3, -5], [8, -7, 6, 4]
    #           E= [1, 2, 3, -5], [-4, 6, 7, 8]
    file_handler = StringIO.StringIO()
    group1_adjs = [(1, 2), (2, 3)]
    group2_adjs = [(6, 7), (7, 8), (3, -5), (7, -8)]
    cars = [[]] + [[i] for i in xrange(1, 9)]
    car_left = [[[], [], [1], [2], [3], [], [5], [6], []],
                [[], [], [1], [2], [3], [4], [5], [6], [7]],
                [[], [], [1], [2], [3], [4], [5], [6], []],
                [[], [], [1], [2], [6], [], [-7], [-6], []],
                [[], [], [1], [2], [-6], [], [-4], [6], [7]]]
    car_right = [[[], [2], [3], [4], [], [6], [7], [-8], [-7]],
                 [[], [2], [3], [4], [5], [6], [7], [8], []],
                 [[], [2], [3], [4], [5], [6], [7], [-8], [-7]],
                 [[], [2], [3], [-5], [], [-3], [4], [-8], []],
                 [[], [2], [3], [-5], [], [-3], [7], [8], []]]
    step_nb = 1
    species1_id = [0, 1, 2]
    species2_id = [3]
    species3_id = [4]
    compute_conserved_adjacencies.find_info_and_write_adjs(group1_adjs, group2_adjs,
                                                           car_left, car_right, cars, step_nb,
                                                           species1_id, species2_id, species3_id,
                                                           file_handler)
    assert file_handler.getvalue() == ("1 2 1 2 0 1 1 1 1 1 1\n2 3 2 3 0 1 1 1 1 1 1\n"
                                       "6 7 6 7 1 1 1 1 1 0 1\n7 8 7 8 1 1 0 1 0 0 1\n"
                                       "3 -5 3 -5 1 1 0 0 0 1 1\n7 -8 7 -8 1 1 1 0 1 1 0\n")


def test_writeretained():
    """
    Test on writing retained adjacencies into a file
    """
    last_adjs = os.path.join("test", "data", "res_S_Adjacencies_file.txt")
    new_adjs = os.path.join("test", "data", "testcomputeconserved.writeretained.txt")
    group1_adjs = [(1, 2), (2, 3)]
    group2_adjs = [(6, 7), (7, 8), (3, -5), (7, -8)]
    cars = [[]] + [[i] for i in xrange(1, 9)]
    car_left = [[[], [], [1], [2], [3], [], [5], [6], []],
                [[], [], [1], [2], [3], [4], [5], [6], [7]],
                [[], [], [1], [2], [3], [4], [5], [6], []],
                [[], [], [1], [2], [6], [], [-7], [-6], []],
                [[], [], [1], [2], [-6], [], [-4], [6], [7]]]
    car_right = [[[], [2], [3], [4], [], [6], [7], [-8], [-7]],
                 [[], [2], [3], [4], [5], [6], [7], [8], []],
                 [[], [2], [3], [4], [5], [6], [7], [-8], [-7]],
                 [[], [2], [3], [-5], [], [-3], [4], [-8], []],
                 [[], [2], [3], [-5], [], [-3], [7], [8], []]]
    step_nb = 1
    species1_id = [0, 1, 2]
    species2_id = [3]
    species3_id = [4]
    compute_conserved_adjacencies.write_retained_adjs(last_adjs, new_adjs, group1_adjs, group2_adjs,
                                                      cars, step_nb, car_left, car_right,
                                                      species1_id, species2_id, species3_id)
    with open(last_adjs, "r") as last:
        lastlines = last.readlines()
    with open(new_adjs, "r") as new:
        newlines = new.readlines()
    assert newlines == ["1 2 1 2 0 1 1 1 1 1 1\n", "2 3 2 3 0 1 1 1 1 1 1\n",
                        "6 7 6 7 1 1 1 1 1 0 1\n", "7 8 7 8 1 1 0 1 0 0 1\n",
                        "3 -5 3 -5 1 1 0 0 0 1 1\n", "7 -8 7 -8 1 1 1 0 1 1 0\n"] + lastlines
    os.remove(new_adjs)


def test_writediscarded():
    """
    Test on writing discarded adjacencies into a file
    """
    filename = os.path.join("test", "data", "plip.txt")
    group1_adjs = [(1, 2), (2, 3)]
    group2_adjs = [(6, 7), (7, 8), (3, -5), (7, -8)]
    cars = [[]] + [[i] for i in xrange(1, 9)]
    car_left = [[[], [], [1], [2], [3], [], [5], [6], []],
                [[], [], [1], [2], [3], [4], [5], [6], [7]],
                [[], [], [1], [2], [3], [4], [5], [6], []],
                [[], [], [1], [2], [6], [], [-7], [-6], []],
                [[], [], [1], [2], [-6], [], [-4], [6], [7]]]
    car_right = [[[], [2], [3], [4], [], [6], [7], [-8], [-7]],
                 [[], [2], [3], [4], [5], [6], [7], [8], []],
                 [[], [2], [3], [4], [5], [6], [7], [-8], [-7]],
                 [[], [2], [3], [-5], [], [-3], [4], [-8], []],
                 [[], [2], [3], [-5], [], [-3], [7], [8], []]]
    step_nb = 1
    species1_id = [0, 1, 2]
    species2_id = [3]
    species3_id = [4]
    compute_conserved_adjacencies.write_discarded_adjs(filename, group1_adjs, group2_adjs,
                                                       car_left, car_right, species1_id,
                                                       species2_id, species3_id, cars, step_nb)
    with open(filename, "r") as res:
        reslines = res.readlines()
    assert reslines == ["1 2 1 2 0 1 1 1 1 1 1\n", "2 3 2 3 0 1 1 1 1 1 1\n",
                        "6 7 6 7 1 1 1 1 1 0 1\n", "7 8 7 8 1 1 0 1 0 0 1\n",
                        "3 -5 3 -5 1 1 0 0 0 1 1\n", "7 -8 7 -8 1 1 1 0 1 1 0\n"]
    os.remove(filename)


def test_find_adjsets(expect_blocks):
    """
    Test on finding all sets of adjacencies according to their status.
    """
    # species tree = ((C,(A,B)),(D,E)@);
    # spe_ids = A:1, B:2, C:0, D: 3, E: 4
    # genomes : A= [1, 2, 3, 4, 5, 6, 7, 8]
    #           B= [1, 2, 3, 4, 5, 6, 7, -8]
    #           C= [1, 2, 3, 4], [5, 6, 7, -8]
    #           D= [1, -2, 3, -5], [8, -7, -6, 4]
    #           E= [1, 2, 3, -5], [-4, 6, 7, 8]
    nb_cars = 8
    nb_species = 5
    blocks = expect_blocks
    cars = [[]] + [[i] for i in xrange(1, 9)]
    blocks_car = [[[], [1, 2, 3, 4], [5, 6, 7, 8]], [[], [1, 2, 3, 4, 5, 6, 7, 8]],
                  [[], [1, 2, 3, 4, 5, 6, 7, 8]], [[], [1, 2, 3, 5], [8, 7, 6, 4]],
                  [[], [1, 2, 3, 5], [4, 6, 7, 8]]]
    block_position_in_car = [0] * 9
    car_neigh_bin = os.path.join("test", "data", "res_neigh_car_information.bin")
    species1_id = [0, 1, 2]
    species2_id = [3]
    species3_id = [4]
    res = compute_conserved_adjacencies.find_adjacency_sets(nb_cars, nb_species, blocks, cars,
                                                            blocks_car, block_position_in_car,
                                                            car_neigh_bin, species1_id,
                                                            species2_id, species3_id)
    car_left, car_right, fs_nc_adj, ps_nc2_adj, fs_c_adj, ps_nc_c_adj, ps_c_adj = res
    assert fs_nc_adj == [[6, 7]]
    assert ps_nc2_adj == [[1, 2], [2, 3], [3, -5], [-4, 6]]
    assert fs_c_adj == []
    assert ps_nc_c_adj == []
    assert ps_c_adj == [[7, 8], [7, -8]]
    with open(car_neigh_bin, "rb") as carfile:
        my_dep = pickle.Unpickler(carfile)
        info = my_dep.load()
    assert info[0] == car_left
    assert info[1] == car_right
