# -*- coding: utf-8 -*-:
"""
Unit tests for compute_initialpqtree.py
"""
from procars.step_modules import compute_initialpqtree
import pytest
import os
try:
    import cPickle as pickle
except ImportError:
    import pickle


# ########################################################################################
# All fixtures needed for tests
@pytest.fixture(scope="module")
def treefile():
    return os.path.join("test", "data", "data_tree.ph")


@pytest.fixture(scope="module")
def blocksfile():
    return os.path.join("test", "data", "data_orthology_blocks.txt")


@pytest.fixture(scope="module")
def expect_tree():
    return {1: ['N1', -1, 2, 7], 2: ['N2', 1, 3, 4], 3: ['C_c', 2, -1, -1], 4: ['N4', 2, 5, 6],
            5: ['A_a', 4, -1, -1], 6: ['B', 4, -1, -1], 7: ['N7', 1, 8, 9],
            8: ['D_species', 7, -1, -1], 9: ['E', 7, -1, -1]}


@pytest.fixture(scope="module")
def expect_blocks():
    return [[[], [[1, 99, 1, 1], [101, 199, 2, 1], [201, 299, 3, 1], [301, 399, 4, 1]],
             [[1, 99, 5, 1], [101, 199, 6, 1], [201, 299, 7, 1], [301, 399, 8, -1]]],
            [[], [[1, 99, 1, 1], [101, 199, 2, 1], [201, 299, 3, 1], [301, 399, 4, 1],
             [401, 499, 5, 1], [501, 599, 6, 1], [601, 699, 7, 1], [701, 799, 8, 1]]],
            [[], [[1, 99, 1, 1], [101, 199, 2, 1], [201, 299, 3, 1], [301, 399, 4, 1],
             [401, 499, 5, 1], [501, 599, 6, 1], [601, 699, 7, 1], [701, 799, 8, -1]]],
            [[], [[1, 99, 1, 1], [101, 199, 2, -1], [201, 299, 3, 1], [301, 399, 5, -1]],
             [[301, 399, 4, 1], [201, 299, 6, -1], [101, 199, 7, -1], [1, 99, 8, 1]]],
            [[], [[1, 99, 1, 1], [101, 199, 2, 1], [201, 299, 3, 1], [301, 399, 5, -1]],
             [[1, 99, 4, -1], [101, 199, 6, 1], [201, 299, 7, 1], [301, 399, 8, 1]]]]


# ########################################################################################
# Unit tests on compute_initialpqtree.py
def test_mark_ancestors(expect_tree):
    """
    Test on marking nodes in the path between root and ancestor
    """
    ancestor_id = 7
    path = compute_initialpqtree.mark_ancestors(expect_tree, ancestor_id)
    assert path == [0, 1, 0, 0, 0, 0, 0, 1, 0, 0]


def test_leaves_internals(expect_tree):
    """
    Test on finding leaves and internal nodes of the tree
    """
    ancestor_id = 7
    path = [0, 1, 0, 0, 0, 0, 0, 1, 0, 0]
    leaves, internals = compute_initialpqtree.list_leaves_internals(expect_tree, ancestor_id, path)
    assert set(leaves) == set([3, 5, 6, 8, 9])
    assert internals == [4, 2, 1, 7]


def test_partition_species(expect_tree):
    """
    Test on partitionning species into three groups: two ingroups and one outgroup, determined
    by the ancestor's position
    """
    species1 = []
    species2 = []
    species3 = []
    ancestor_id = 7
    cur_id = 1
    partition_id = 1
    spe1, spe2, spe3 = compute_initialpqtree.partition_species(species1, species2, species3,
                                                               expect_tree, ancestor_id, cur_id,
                                                               partition_id)
    assert set(spe1) == set(['A_a', 'B', 'C_c'])
    assert spe2 == ['D_species']
    assert spe3 == ["E"]


def test_compute_partition_species(expect_tree):
    """
    Test on finding IDs of species partition found
    """
    ancestor_id = 7
    res = compute_initialpqtree.compute_partition_and_species_id(expect_tree, ancestor_id)
    spe1, spe2, spe3, spe_ids = res
    assert set(spe1) == set([0, 1, 2])
    assert spe2 == [3]
    assert spe3 == [4]
    assert spe_ids == {'C_c': 0, 'B': 2, 'E': 4, 'D_species': 3, 'A_a': 1}


def test_compute_params_tree(treefile, expect_tree):
    """
    Test on computing all parameters from a tree file
    """
    res = compute_initialpqtree.compute_parameters_from_tree(treefile)
    assert res[0] == expect_tree
    assert res[1] == 7
    assert res[2] == [0, 1, 0, 0, 0, 0, 0, 1, 0, 0]
    assert set(res[3]) == set([3, 5, 6, 8, 9])
    assert res[4] == [4, 2, 1, 7]


def test_commons_read_tree(treefile, expect_tree):
    """
    Test on reading a tree file, getting all info, and partitionning according to ancestor id
    """
    binary_tree_file = os.path.join("test", "data", "tree_information.bin")
    ref_binfile = os.path.join("test", "data", "res_tree_information.bin")
    res = compute_initialpqtree.commons_read_tree(treefile, binary_tree_file)
    with open(binary_tree_file, "rb") as treefile:
        my_dep = pickle.Unpickler(treefile)
        info = my_dep.load()
    with open(ref_binfile, "rb") as reftreefile:
        my_dep = pickle.Unpickler(reftreefile)
        info_ref = my_dep.load()
    for index, inf in enumerate(info):
        if index == 3:
            assert set(inf) == set(info_ref[index])
        else:
            assert inf == info_ref[index]
    os.remove(binary_tree_file)


def test_compute_blocks_neigh(expect_blocks):
    """
    Test on finding previous and next blocks for each block of each species
    """
    nb_blocks = 8
    nb_species = 5
    left, right = compute_initialpqtree.compute_blocks_neighbors(nb_blocks, nb_species,
                                                                 expect_blocks)
    assert left == [[[0, 0], [0, 0], [0, 1], [0, 2], [0, 3], [0, 0], [0, 5], [0, 6], [0, 0]],
                    [[0, 0], [0, 0], [0, 1], [0, 2], [0, 3], [0, 4], [0, 5], [0, 6], [0, 7]],
                    [[0, 0], [0, 0], [0, 1], [0, 2], [0, 3], [0, 4], [0, 5], [0, 6], [0, 0]],
                    [[0, 0], [0, 0], [0, -3], [0, -2], [0, -6], [0, 0], [0, -4], [0, 6], [0, 0]],
                    [[0, 0], [0, 0], [0, 1], [0, 2], [0, -6], [0, 0], [0, -4], [0, 6], [0, 7]]]
    assert right == [[[0], [0, 2], [0, 3], [0, 4], [0, 0], [0, 6], [0, 7], [0, -8], [0, -7]],
                     [[0], [0, 2], [0, 3], [0, 4], [0, 5], [0, 6], [0, 7], [0, 8], [0, 0]],
                     [[0], [0, 2], [0, 3], [0, 4], [0, 5], [0, 6], [0, 7], [0, -8], [0, -7]],
                     [[0], [0, -2], [0, -1], [0, -5], [0, 0], [0, -3], [0, 7], [0, -8], [0, -7]],
                     [[0], [0, 2], [0, 3], [0, -5], [0, 0], [0, -3], [0, 7], [0, 8], [0, 0]]]


def test_commons_read_blocks(blocksfile, expect_blocks):
    """
    Test on reading a block file, and finding all previous and next blocks
    """
    nb_species = 5
    spe_ids = {'C_c': 0, 'B': 2, 'E': 4, 'D_species': 3, 'A_a': 1}
    binfile = os.path.join("test", "data", "block_information.bin")
    ref_binfile = os.path.join("test", "data", "res_block_information.bin")
    compute_initialpqtree.commons_read_blocks(blocksfile, nb_species, spe_ids, binfile)
    with open(binfile, "rb") as blockfile:
        my_dep = pickle.Unpickler(blockfile)
        info = my_dep.load()
    with open(ref_binfile, "rb") as refblockfile:
        my_dep = pickle.Unpickler(refblockfile)
        info_ref = my_dep.load()
    for index, inf in enumerate(info):
        assert inf == info_ref[index]
    os.remove(binfile)


def test_write_cars():
    """
    Test on writing the initial PQtree (one CAR per block)
    """
    nb_blocks = 8
    car_file_name = os.path.join("test", "data", "PQtree0.txt")
    ref_carfile = os.path.join("test", "data", "res_PQtree_0.txt")
    compute_initialpqtree.write_init_cars_file(car_file_name, nb_blocks)
    with open(car_file_name, "r") as carfile:
        cars = carfile.readlines()
    with open(ref_carfile, "r") as refcarfile:
        ref_cars = refcarfile.readlines()
    assert cars == ref_cars
    os.remove(car_file_name)
