# -*- coding: utf-8 -*-:
"""
Unit tests for compute_pqtree.py
"""

from procars.step_modules import compute_pqtree
import pytest
import os
import shutil


def test_readfromblock_left():
    """
    Test on finding a car from the left of a given block
    """
    end = 0
    left = {1: 0, 2: -5, 3: -4, 4: -3, 5: -2, 6: 0, 7: 4, 8: 0, 9: 2}
    right = {1: -3, 2: 9, 3: -1, 4: 7, 5: -8, 6: 0, 7: -9, 8: -5, 9: -7}
    signed_block = 7
    car = compute_pqtree.read_from_block(signed_block, left, right, end)
    assert car == [1, -3, 4, 7]


def test_readfromblock_right():
    """
    Test on finding a car from the right of a given block
    """
    end = 1
    left = {1: 0, 2: -5, 3: -4, 4: -3, 5: -2, 6: 0, 7: 4, 8: 0, 9: 2}
    right = {1: -3, 2: 9, 3: -1, 4: 7, 5: -8, 6: 0, 7: -9, 8: -5, 9: -7}
    signed_block = -9
    car = compute_pqtree.read_from_block(signed_block, left, right, end)
    assert car == [-9, -2, 5, -8]


def test_readfromblock_cycle():
    """
    Test on finding a car from the right of a given block in a cycle car
    """
    end = 1
    left = {1: -8, 2: -5, 3: -4, 4: -3, 5: -2, 6: 0, 7: 4, 8: -1, 9: 2}
    right = {1: -3, 2: 9, 3: -1, 4: 7, 5: -8, 6: 0, 7: -9, 8: -5, 9: -7}
    signed_block = -7
    car = compute_pqtree.read_from_block(signed_block, left, right, end)
    assert car == []


def test_readfromblock_0():
    """
    Test on finding a car from the right of block 0 (does not exist)
    """
    end = 1
    left = {1: -8, 2: -5, 3: -4, 4: -3, 5: -2, 6: 0, 7: 4, 8: -1, 9: 2}
    right = {1: -3, 2: 9, 3: -1, 4: 7, 5: -8, 6: 0, 7: -9, 8: -5, 9: -7}
    signed_block = 0
    car = compute_pqtree.read_from_block(signed_block, left, right, end)
    assert car == []


def test_init_readcycle_discardunresolved():
    """
    Test on initialization of resolution of cycle CARs
    car should be 1 -3 4 7 -9 -2 5 -8
    with (7 -9) just added, and in discarded
    """
    signed_block = 7
    end = 1
    nex_signed_block = -9
    next_prev_signed_block = 0
    next_prev_discarded_signed_block = [8, -9]
    res = compute_pqtree.init_read_from_block_in_cycle(signed_block, nex_signed_block,
                                                       next_prev_signed_block,
                                                       next_prev_discarded_signed_block, end)
    to_be_discarded_unresolved, to_be_discarded_resolved, current_signed_block = res
    assert to_be_discarded_unresolved == [[7, -9]]
    assert to_be_discarded_resolved == []
    assert current_signed_block == -9


def test_init_readcycle_discardresolved():
    """
    Test on initialization of resolution of cycle CARs
    CARs should be 1 -3 4 7 and -9 -2 5 -8
    with (7 -9) just added, and not in discarded
    """
    signed_block = -7
    end = 0
    nex_signed_block = 9
    next_prev_signed_block = 0
    next_prev_discarded_signed_block = [8]
    res = compute_pqtree.init_read_from_block_in_cycle(signed_block, nex_signed_block,
                                                       next_prev_signed_block,
                                                       next_prev_discarded_signed_block, end)
    to_be_discarded_unresolved, to_be_discarded_resolved, current_signed_block = res
    assert to_be_discarded_unresolved == []
    assert to_be_discarded_resolved == [[9, -7]]
    assert current_signed_block == 0


def test_init_readcycle_nodiscard():
    """
    Test on initialization of resolution of cycle CARs
    CARs should be 1 -3 4 7 and -9 -2 5 -8
    with (7 -9) just added, and not in discarded
    """
    signed_block = 4
    end = 1
    nex_signed_block = 7
    next_prev_signed_block = 7
    next_prev_discarded_signed_block = []
    res = compute_pqtree.init_read_from_block_in_cycle(signed_block, nex_signed_block,
                                                       next_prev_signed_block,
                                                       next_prev_discarded_signed_block, end)
    to_be_discarded_unresolved, to_be_discarded_resolved, current_signed_block = res
    assert to_be_discarded_unresolved == []
    assert to_be_discarded_resolved == []
    assert current_signed_block == 7


def test_readfromblockincycle_left():
    """
    Test on finding a linear car from the left of a given block, given that if we add all
    adjacencies we get a cycle car -> remove less reliable adjacency
    Cycle is resolved: car = 1 -3 4 7 -9 -2 5 -8
    remove adjacency -8 1
    """
    signed_block = 7
    end = 0
    left = {1: -8, 2: -5, 3: -4, 4: -3, 5: -2, 6: 0, 7: 4, 8: -1, 9: 2}
    right = {1: -3, 2: 9, 3: -1, 4: 7, 5: -8, 6: 0, 7: -9, 8: -5, 9: -7}
    left_prev = {1: 0, 2: -5, 3: -4, 4: -3, 5: -2, 6: 0, 7: 4, 8: 0, 9: 2}
    right_prev = {1: -3, 2: 9, 3: -1, 4: 7, 5: -8, 6: 0, 7: 0, 8: -5, 9: 0}
    left_prev_discarded = {1: [], 2: [], 3: [], 4: [], 5: [], 6: [], 7: [], 8: [7], 9: []}
    right_prev_discarded = {1: [], 2: [], 3: [], 4: [], 5: [], 6: [], 7: [-9, 8], 8: [], 9: [-7]}
    res = compute_pqtree.read_from_block_in_cycle(signed_block, left, right, left_prev, right_prev,
                                                  left_prev_discarded, right_prev_discarded, end)
    car, to_be_discarded = res
    assert to_be_discarded == [[-8, 1]]
    assert car == [1, -3, 4, 7]


def test_readfromblockincycle_right():
    """
    Test on finding a linear car from the right of a given block, given that if we add all
    adjacencies we get a cycle car -> remove less reliable adjacency
    Cycle is resolved: car = 1 -3 4 7 -9 -2 5 -8
    remove adjacency -8 1
    """
    signed_block = -9
    end = 1
    left = {1: -8, 2: -5, 3: -4, 4: -3, 5: -2, 6: 0, 7: 4, 8: -1, 9: 2}
    right = {1: -3, 2: 9, 3: -1, 4: 7, 5: -8, 6: 0, 7: -9, 8: -5, 9: -7}
    left_prev = {1: 0, 2: -5, 3: -4, 4: -3, 5: -2, 6: 0, 7: 4, 8: 0, 9: 2}
    right_prev = {1: -3, 2: 9, 3: -1, 4: 7, 5: -8, 6: 0, 7: 0, 8: -5, 9: 0}
    left_prev_discarded = {1: [], 2: [], 3: [], 4: [], 5: [], 6: [], 7: [], 8: [7], 9: []}
    right_prev_discarded = {1: [], 2: [], 3: [], 4: [], 5: [], 6: [], 7: [-9, 8], 8: [], 9: [-7]}
    res = compute_pqtree.read_from_block_in_cycle(signed_block, left, right, left_prev, right_prev,
                                                  left_prev_discarded, right_prev_discarded, end)
    car, to_be_discarded = res
    assert to_be_discarded == [[-8, 1]]
    assert car == [-9, -2, 5, -8]


def test_readfromblockincycle_unresolved():
    """
    Test on finding a linear car from the left of a given block, given that if we add all
    adjacencies we get a cycle car: all adjacencies have the same weight
    (are in previously discarded adjs), we cannot resolve the cycle.
    cars = 1 -3 4 7 and -9 -2 5 -8 as previously
    -> remove adjacencies 7 -9 and -8 1
    """
    signed_block = -9
    end = 0
    left = {1: -8, 2: -5, 3: -4, 4: -3, 5: -2, 6: 0, 7: 4, 8: -1, 9: 2}
    right = {1: -3, 2: 9, 3: -1, 4: 7, 5: -8, 6: 0, 7: -9, 8: -5, 9: -7}
    left_prev = {1: 0, 2: -5, 3: -4, 4: -3, 5: -2, 6: 0, 7: 4, 8: 0, 9: 2}
    right_prev = {1: -3, 2: 9, 3: -1, 4: 7, 5: -8, 6: 0, 7: 0, 8: -5, 9: 0}
    left_prev_discarded = {1: [-8, 9], 2: [], 3: [], 4: [], 5: [], 6: [], 7: [], 8: [7, -1], 9: []}
    right_prev_discarded = {1: [], 2: [], 3: [], 4: [], 5: [], 6: [], 7: [8, -9], 8: [], 9: [-7, 1]}
    res = compute_pqtree.read_from_block_in_cycle(signed_block, left, right, left_prev, right_prev,
                                                  left_prev_discarded, right_prev_discarded, end)
    car, to_be_discarded = res
    assert to_be_discarded == [[7, -9], [-8, 1]]
    assert car == [-9, -2, 5, -8, 1, -3, 4, 7, -9]


def test_readfromblockincycle_unresolved_neg():
    """
    Test on finding a linear car from the left of a given block, given that if we add all
    adjacencies we get a cycle car: all adjacencies have the same weight
    (are in previously discarded adjs), we cannot resolve the cycle.
    cars = 1 -3 4 7 and -9 -2 5 -8 as previously
    -> remove adjacencies 7 -9 and -8 1
    """
    signed_block = -2
    end = 0
    left = {1: -8, 2: -5, 3: -4, 4: -3, 5: -2, 6: 0, 7: 4, 8: -1, 9: 2}
    right = {1: -3, 2: 9, 3: -1, 4: 7, 5: -8, 6: 0, 7: -9, 8: -5, 9: -7}
    left_prev = {1: 0, 2: -5, 3: -4, 4: -3, 5: -2, 6: 0, 7: 4, 8: 0, 9: 2}
    right_prev = {1: -3, 2: 9, 3: -1, 4: 7, 5: -8, 6: 0, 7: 0, 8: -5, 9: 0}
    left_prev_discarded = {1: [-8, 9], 2: [], 3: [], 4: [], 5: [], 6: [], 7: [], 8: [7, -1], 9: []}
    right_prev_discarded = {1: [], 2: [], 3: [], 4: [], 5: [], 6: [], 7: [8, -9], 8: [], 9: [-7, 1]}
    res = compute_pqtree.read_from_block_in_cycle(signed_block, left, right, left_prev, right_prev,
                                                  left_prev_discarded, right_prev_discarded, end)
    car, to_be_discarded = res
    assert to_be_discarded == [[7, -9], [-8, 1]]
    assert car == [-2, 5, -8, 1, -3, 4, 7, -9, -2]


def test_readfromblockincycle_unresolved_neg_right():
    """
    Test on finding a linear car from the left of a given block, given that if we add all
    adjacencies we get a cycle car: all adjacencies have the same weight
    (are in previously discarded adjs), we cannot resolve the cycle.
    cars = 1 -3 4 7 and -9 -2 5 -8 as previously
    -> remove adjacencies 7 -9 and -8 1
    """
    signed_block = -2
    end = 1
    left = {1: -8, 2: -5, 3: -4, 4: -3, 5: -2, 6: 0, 7: 4, 8: -1, 9: 2}
    right = {1: -3, 2: 9, 3: -1, 4: 7, 5: -8, 6: 0, 7: -9, 8: -5, 9: -7}
    left_prev = {1: 0, 2: -5, 3: -4, 4: -3, 5: -2, 6: 0, 7: 4, 8: 0, 9: 2}
    right_prev = {1: -3, 2: 9, 3: -1, 4: 7, 5: -8, 6: 0, 7: 0, 8: -5, 9: 0}
    left_prev_discarded = {1: [-8, 9], 2: [], 3: [], 4: [], 5: [], 6: [], 7: [], 8: [7, -1], 9: []}
    right_prev_discarded = {1: [], 2: [], 3: [], 4: [], 5: [], 6: [], 7: [8, -9], 8: [], 9: [-7, 1]}
    res = compute_pqtree.read_from_block_in_cycle(signed_block, left, right, left_prev, right_prev,
                                                  left_prev_discarded, right_prev_discarded, end)
    car, to_be_discarded = res
    assert to_be_discarded == [[-8, 1], [7, -9]]
    assert car == [-2, 5, -8, 1, -3, 4, 7, -9, -2]


def test_readfromblockincycle_2cars():
    """
    Test on finding a linear car from the left of a given block, given that if we add all
    adjacencies we get a cycle car: any of the adjacencies were present at the last step (nor
    discarded): do not change anything, keep last 2 cars instead of making a cycle
    -> cars = 1 -3 4 7 and -9 -2 5 -8 as previously
    -> remove adjacencies 7 -9 and -8 1
    """
    signed_block = -2
    end = 0
    left = {1: -8, 2: -5, 3: -4, 4: -3, 5: -2, 6: 0, 7: 4, 8: -1, 9: 2}
    right = {1: -3, 2: 9, 3: -1, 4: 7, 5: -8, 6: 0, 7: -9, 8: -5, 9: -7}
    left_prev = {1: 0, 2: -5, 3: -4, 4: -3, 5: -2, 6: 0, 7: 4, 8: 0, 9: 2}
    right_prev = {1: -3, 2: 9, 3: -1, 4: 7, 5: -8, 6: 0, 7: 0, 8: -5, 9: 0}
    left_prev_discarded = {1: [], 2: [], 3: [], 4: [], 5: [], 6: [], 7: [], 8: [], 9: []}
    right_prev_discarded = {1: [], 2: [], 3: [], 4: [], 5: [], 6: [], 7: [], 8: [], 9: []}
    res = compute_pqtree.read_from_block_in_cycle(signed_block, left, right, left_prev, right_prev,
                                                  left_prev_discarded, right_prev_discarded, end)
    car, to_be_discarded = res
    print to_be_discarded == [[7, -9], [-8, 1]]
    assert car == [-9, -2]


def test_remove_adjinfile():
    """
    Test on removing a given adjacency (same order as it is written in file) from the adjacency file
    """
    refadj = os.path.join("test", "data", "res_S_Adjacencies_file.txt")
    lastadj = os.path.join("test", "data", "testcomputepqtree.removeadjfromfile.txt")
    shutil.copyfile(refadj, lastadj)
    adjacency = (-17, 25)
    compute_pqtree.remove_adjacency_from_file(lastadj, adjacency)
    with open(refadj, "r") as ref:
        reflines = ref.readlines()
    with open(lastadj, "r") as last:
        lastlines = last.readlines()
    assert set(reflines) == set(lastlines + ["-17 25 2 -4 1 2\n"])


def test_remove_adjinfile_rev():
    """
    Test on removing a given adjacency (reverse order than it is written in file)
    from the adjacency file
    """
    refadj = os.path.join("test", "data", "res_S_Adjacencies_file.txt")
    lastadj = os.path.join("test", "data", "testcomputepqtree.removeadjfromfile.txt")
    shutil.copyfile(refadj, lastadj)
    adjacency = (-25, 17)
    compute_pqtree.remove_adjacency_from_file(lastadj, adjacency)
    with open(refadj, "r") as ref:
        reflines = ref.readlines()
    with open(lastadj, "r") as last:
        lastlines = last.readlines()
    assert set(reflines) == set(lastlines + ["-17 25 2 -4 1 2\n"])
    os.remove(lastadj)


def test_findcars_nocycle():
    """
    Test on finding all cars from the list of adjacencies. No cycle car.
    """
    left = {1: 0, 2: -5, 3: -4, 4: -3, 5: -2, 6: 0, 7: 4, 8: 0, 9: 2}
    right = {1: -3, 2: 9, 3: -1, 4: 7, 5: -8, 6: 0, 7: -9, 8: -5, 9: -7}
    adj_prefix = os.path.join("test", "data", "res_testcomputepqtree_findcars")
    step_nb = 3
    nb_blocks = 9
    cars = compute_pqtree.find_cars(nb_blocks, left, right, adj_prefix, step_nb)
    assert cars == [[1, -3, 4, 7, -9, -2, 5, -8], [6]]


def test_findcars_cycleresolved():
    """
    Test on finding all cars from the list of adjacencies. Cycle car resolved thanks to discarded
    adjacencies at last step. -> remove one adjacency from current step adj file.
    """
    left = {1: -8, 2: -5, 3: -4, 4: -3, 5: -2, 6: 0, 7: 4, 8: -1, 9: 2}
    right = {1: -3, 2: 9, 3: -1, 4: 7, 5: -8, 6: 0, 7: -9, 8: -5, 9: -7}
    adj_prefix = os.path.join("test", "data", "res_testcomputepqtree_findcars")
    shutil.copyfile(adj_prefix + "_3ref.txt", adj_prefix + "_3.txt")
    step_nb = 3
    nb_blocks = 9
    cars = compute_pqtree.find_cars(nb_blocks, left, right, adj_prefix, step_nb)
    assert cars == [[1, -3, 4, 7, -9, -2, 5, -8], [6]]
    with open(adj_prefix + "_3.txt", "r") as res:
        lines = res.readlines()
    with open(adj_prefix + "_3ref.txt", "r") as ref:
        reflines = ref.readlines()
    assert set(reflines) == set(lines + ["-8 1 2 3 1 3\n"])
    os.remove(adj_prefix + "_3.txt")


def test_findcars_cycleunresolved():
    """
    Test on finding all cars from the list of adjacencies. Cycle CAR not resolved because the
    two adjs forming the cycle were present in discarded adjs.
    """
    left = {1: -8, 2: -5, 3: -4, 4: -3, 5: -2, 6: 0, 7: 4, 8: -1, 9: 2}
    right = {1: -3, 2: 9, 3: -1, 4: 7, 5: -8, 6: 0, 7: -9, 8: -5, 9: -7}
    adj_prefix = os.path.join("test", "data", "res_testcomputepqtree_findcars_unres")
    refadjlast = os.path.join("test", "data", "res_testcomputepqtree_findcars_2.txt")
    refadjnew = os.path.join("test", "data", "res_testcomputepqtree_findcars_3ref.txt")
    shutil.copyfile(refadjlast, adj_prefix + "_2.txt")
    shutil.copyfile(refadjnew, adj_prefix + "_3.txt")
    step_nb = 3
    nb_blocks = 9
    cars = compute_pqtree.find_cars(nb_blocks, left, right, adj_prefix, step_nb)
    assert cars == [[1, -3, 4, 7], [-9, -2, 5, -8], [6]]
    with open(adj_prefix + "_3.txt", "r") as res:
        lines = res.readlines()
    with open(refadjnew, "r") as ref:
        reflines = ref.readlines()
    assert set(reflines) == set(lines + ["-8 1 2 3 1 3\n", "7 -9 1 2 1 3\n"])
    os.remove(adj_prefix + "_3.txt")
    os.remove(adj_prefix + "_2.txt")


def test_findcars_cycleunresolved2():
    """
    Test on finding all cars from the list of adjacencies. Cycle car not resolved because no
    adjacency forming the cycle was present in discarded adjacencies.
    """
    left = {1: -8, 2: -5, 3: -4, 4: -3, 5: -2, 6: 0, 7: 4, 8: -1, 9: 2}
    right = {1: -3, 2: 9, 3: -1, 4: 7, 5: -8, 6: 0, 7: -9, 8: -5, 9: -7}
    adj_prefix = os.path.join("test", "data", "res_testcomputepqtree_findcars_unres2")
    refadjlast = os.path.join("test", "data", "res_testcomputepqtree_findcars_2.txt")
    refadjnew = os.path.join("test", "data", "res_testcomputepqtree_findcars_3ref.txt")
    shutil.copyfile(refadjlast, adj_prefix + "_2.txt")
    shutil.copyfile(refadjnew, adj_prefix + "_3.txt")
    # empty discarded adjacencies file for last step
    open(adj_prefix + "_2.txt_discarded.txt", "w").close()
    step_nb = 3
    nb_blocks = 9
    cars = compute_pqtree.find_cars(nb_blocks, left, right, adj_prefix, step_nb)
    print cars
    assert cars == [[1, -3, 4, 7], [8, -5, 2, 9], [6]]
    with open(adj_prefix + "_3.txt", "r") as res:
        lines = res.readlines()
    with open(refadjnew, "r") as ref:
        reflines = ref.readlines()
    assert set(reflines) == set(lines + ["-8 1 2 3 1 3\n", "7 -9 1 2 1 3\n"])
    os.remove(adj_prefix + "_3.txt")
    os.remove(adj_prefix + "_2.txt")
    os.remove(adj_prefix + "_2.txt_discarded.txt")
