# -*- coding: utf-8 -*-:
"""
Unit tests for compute_dcj_adjacencies.py
"""


from procars.utils.util_classes import SparseMatrix
from procars.step_modules import compute_dcj_adjacencies
import pytest
import os
try:
    import cPickle as pickle
except ImportError:
    import pickle


def test_compute_neighbor_leftpos():
    """
    Test on finding a positive block's left neighbor
    """
    spe = 1
    block = 3
    left = [[], [[], [0, 0], [0, -5], [0, -4], [0, -3], [0, -2], [0, 0], [0, 4], [0, 0], [0, 2]]]
    right = [[], [[], [0, -3], [0, 9], [0, -1], [0, -1], [0, 7], [0, -8], [0, 0], [0, -9],
             [0, -5], [0, -7]]]
    side = 0
    neigh = compute_dcj_adjacencies.compute_neighbor(spe, block, side, left, right)
    assert neigh == -4


def test_compute_neighbor_leftneg():
    """
    Test on finding a negative block's left neighbor
    """
    spe = 1
    block = -3
    left = [[], [[], [0, 0], [0, -5], [0, -4], [0, -3], [0, -2], [0, 0], [0, 4], [0, 0], [0, 2]]]
    right = [[], [[], [0, -3], [0, 9], [0, -1], [0, -1], [0, 7], [0, -8], [0, 0], [0, -9],
             [0, -5], [0, -7]]]
    side = 0
    neigh = compute_dcj_adjacencies.compute_neighbor(spe, block, side, left, right)
    assert neigh == 1


def test_compute_neighbor_rightpos():
    """
    Test on finding a positive block's right neighbor
    """
    spe = 1
    block = 4
    left = [[], [[], [0, 0], [0, -5], [0, -4], [0, -3], [0, -2], [0, 0], [0, 4], [0, 0], [0, 2]]]
    right = [[], [[], [0, -3], [0, 9], [0, -1], [0, 7], [0, -8], [0, 0], [0, -9],
             [0, -5], [0, -7]]]
    side = 1
    neigh = compute_dcj_adjacencies.compute_neighbor(spe, block, side, left, right)
    assert neigh == 7


def test_compute_neighbor_rightneg():
    """
    Test on finding a negative block's right neighbor
    """
    spe = 1
    block = -4
    left = [[], [[], [0, 0], [0, -5], [0, -4], [0, -3], [0, -2], [0, 0], [0, 4], [0, 0], [0, 2]]]
    right = [[], [[], [0, -3], [0, 9], [0, -1], [0, 7], [0, -8], [0, 0], [0, -9],
             [0, -5], [0, -7]]]
    side = 1
    neigh = compute_dcj_adjacencies.compute_neighbor(spe, block, side, left, right)
    assert neigh == 3


def test_isadj():
    """
    Test that a given adjacency belongs to cars
    """
    adj = (-7, -4)
    cars = [[1, -3, 4, 7], [6], [-9, -2, 5, -8]]
    assert compute_dcj_adjacencies.is_adjacency_in_car(cars, adj)


def test_isnotadj():
    """
    Test that a given adjacency does not belong to cars
    """
    adj = (-4, 7)
    cars = [[1, -3, 4, 7], [6], [-9, -2, 5, -8]]
    assert not compute_dcj_adjacencies.is_adjacency_in_car(cars, adj)


def test_iscaradj():
    """
    Test that a given car adj is possible
    """
    spe = 1
    car_adj = (-2, -1)
    car_left = [[], [[], [], [1, -1], [2, -1]]]
    car_right = [[], [[], [2], [], [-1]]]
    assert compute_dcj_adjacencies.is_car_adjacency(spe, car_adj, car_left, car_right)


def test_isnotcaradj():
    """
    Test that a given car adj is not possible
    """
    spe = 1
    car_adj = (1, 3)
    car_left = [[], [[], [], [1, -1], [2, -1]]]
    car_right = [[], [[], [2], [], [-1]]]
    assert not compute_dcj_adjacencies.is_car_adjacency(spe, car_adj, car_left, car_right)


def test_dcjreliable_conserved():
    """
    Test that a car adjacency is a dcj reliable conserved in the species tree
    """
    # Species tree: (A, (B, C)@);
    # IDs : A=1, B=2, C=3
    # genomes : A: 4 3 and 1 -2
    #           B: 1 -4 and 2 3
    #           C: 1 2 -3 4
    # block adj (1, -4) or car adj (2, 3) is a dcj-reliable adjacency
    partition = [[1], [2], [3]]
    car_adj = (2, 3)
    block_adj = (1, -4)
    cars = [[], [2, 3], [5, 1], [-4, 6]]
    left = [[], [[], [0, 0], [0, 0], [0, 4], [0, 0]],
            [[], [0, 0], [0, 0], [0, 2], [0, 0]],
            [[], [0, 0], [0, 1], [0, -4], [0, -3]]]
    right = [[], [[], [0, -2], [0, -1], [0, 0], [0, 3]],
             [[], [0, -4], [0, 3], [0, 0], [0, -1]],
             [[], [0, 2], [0, -3], [0, -2], [0, 0]]]
    car_left = [[], [[], [], [], []], [[], [], [], [2]], [[], [], [], []]]
    car_right = [[], [[], [], [], []], [[], [], [3], []], [[], [], [], []]]
    an = compute_dcj_adjacencies.is_dcj_reliable_conserved(partition, car_adj, block_adj, cars,
                                                           left, right, car_left, car_right)
    assert an


def test_dcjreliable_notconserved():
    """
    Test that a car adjacency is a dcj reliable conserved in the species tree
    """
    # Species tree: (A, (B, C)@);
    # IDs : A=1, B=2, C=3
    # genomes : A: 4 3 and 1 -2
    #           B: 1 -4 and 2 3
    #           C: 1 2 -3 4
    # block adj (3, 1) or car adj (1, 2) is NOT a dcj-reliable adjacency
    # (no signal in any extant genome)
    partition = [[1], [2], [3]]
    car_adj = (1, 2)
    block_adj = (3, 1)
    cars = [[], [2, 3], [1], [-4]]
    left = [[], [[], [0, 0], [0, 0], [0, 4], [0, 0]],
            [[], [0, 0], [0, 0], [0, 2], [0, 0]],
            [[], [0, 0], [0, 1], [0, -4], [0, -3]]]
    right = [[], [[], [0, -2], [0, -1], [0, 0], [0, 3]],
             [[], [0, -4], [0, 3], [0, 0], [0, -1]],
             [[], [0, 2], [0, -3], [0, -2], [0, 0]]]
    car_left = [[], [[], [], [], []], [[], [], [], [2]], [[], [], [], []]]
    car_right = [[], [[], [], [], []], [[], [], [3], []], [[], [], [], []]]
    an = compute_dcj_adjacencies.is_dcj_reliable_conserved(partition, car_adj, block_adj, cars,
                                                           left, right, car_left, car_right)
    assert not an


def test_dcjreliable():
    """
    Tests if a given adjacency is dcj-reliable in any of the possible partitions
    """
    # Species tree: (A, (B, C)@);
    # IDs : A=1, B=2, C=3
    # genomes : A: 4 3 and 1 -2
    #           B: 1 -4 and 2 3
    #           C: 1 2 -3 4
    # block adj (1, -4) or car adj (2, 3) is a dcj-reliable adjacency
    possible_partitions = [[[2], [3], [1]], [[3], [2], [1]], [[1], [2], [3]]]
    car_adjacency = (2, 3)
    cars = [[], [2, 3], [5, 1], [-4, 6]]
    left = [[], [[], [0, 0], [0, 0], [0, 4], [0, 0]],
            [[], [0, 0], [0, 0], [0, 2], [0, 0]],
            [[], [0, 0], [0, 1], [0, -4], [0, -3]]]
    right = [[], [[], [0, -2], [0, -1], [0, 0], [0, 3]],
             [[], [0, -4], [0, 3], [0, 0], [0, -1]],
             [[], [0, 2], [0, -3], [0, -2], [0, 0]]]
    car_left = [[], [[], [], [], []], [[], [], [], [2]], [[], [], [], []]]
    car_right = [[], [[], [], [], []], [[], [], [3], []], [[], [], [], []]]
    an = compute_dcj_adjacencies.is_dcj_reliable(possible_partitions, car_adjacency, cars, left,
                                                 right, car_left, car_right)
    assert an


def test_dcjreliable_neg():
    """
    Tests if a given adjacency is dcj-reliable in any of the possible partitions
    """
    # Species tree: (A, (B, C)@);
    # IDs : A=1, B=2, C=3
    # genomes : A: 4 3 and 1 -2
    #           B: 1 -4 and 2 3
    #           C: 1 2 -3 4
    # block adj (1, -4) or car adj (2, 3) is a dcj-reliable adjacency
    possible_partitions = [[[2], [3], [1]], [[3], [2], [1]], [[1], [2], [3]]]
    car_adjacency = (2, -3)
    cars = [[], [2, 3], [1], [-4]]
    left = [[], [[], [0, 0], [0, 0], [0, 4], [0, 0]],
            [[], [0, 0], [0, 0], [0, 2], [0, 0]],
            [[], [0, 0], [0, 1], [0, -4], [0, -3]]]
    right = [[], [[], [0, -2], [0, -1], [0, 0], [0, 3]],
             [[], [0, -4], [0, 3], [0, 0], [0, -1]],
             [[], [0, 2], [0, -3], [0, -2], [0, 0]]]
    car_left = [[], [[], [], [], []], [[], [], [], [2]], [[], [], [], []]]
    car_right = [[], [[], [], [], []], [[], [], [3], []], [[], [], [], []]]
    an = compute_dcj_adjacencies.is_dcj_reliable(possible_partitions, car_adjacency, cars, left,
                                                 right, car_left, car_right)
    assert not an


def test_find_dcjs():
    """
    Test on finding possible dcj adjacencies from a set of cars
    """
    nb_cars = 3
    possible_partitions = [[[2], [3], [1]], [[3], [2], [1]], [[1], [2], [3]]]
    cars = [[], [2, 3], [1], [-4]]
    left = [[], [[], [0, 0], [0, 0], [0, 4], [0, 0]],
            [[], [0, 0], [0, 0], [0, 2], [0, 0]],
            [[], [0, 0], [0, 1], [0, -4], [0, -3]]]
    right = [[], [[], [0, -2], [0, -1], [0, 0], [0, 3]],
             [[], [0, -4], [0, 3], [0, 0], [0, -1]],
             [[], [0, 2], [0, -3], [0, -2], [0, 0]]]
    car_left = [[], [[], [], [], []], [[], [], [], [2]], [[], [], [], []]]
    car_right = [[], [[], [], [], []], [[], [], [3], []], [[], [], [], []]]
    dcj_adjs = compute_dcj_adjacencies.find_possible_dcj_adjs(nb_cars, possible_partitions, cars,
                                                              left, right, car_left, car_right)
    assert dcj_adjs == [[2, 3]]


def test_find_dcjs_negcars():
    """
    Test on finding possible dcj adjacencies from a set of cars
    """
    nb_cars = 3
    possible_partitions = [[[2], [3], [1]], [[3], [2], [1]], [[1], [2], [3]]]
    cars = [[], [2, 3], [1], [4]]
    left = [[], [[], [0, 0], [0, 0], [0, 4], [0, 0]],
            [[], [0, 0], [0, 0], [0, 2], [0, 0]],
            [[], [0, 0], [0, 1], [0, -4], [0, -3]]]
    right = [[], [[], [0, -2], [0, -1], [0, 0], [0, 3]],
             [[], [0, -4], [0, 3], [0, 0], [0, -1]],
             [[], [0, 2], [0, -3], [0, -2], [0, 0]]]
    car_left = [[], [[], [], [], []], [[], [], [], []], [[], [], [], []]]
    car_right = [[], [[], [], [], []], [[], [], [-3], [-2]], [[], [], [], []]]
    dcj_adjs = compute_dcj_adjacencies.find_possible_dcj_adjs(nb_cars, possible_partitions, cars,
                                                              left, right, car_left, car_right)
    assert dcj_adjs == [[2, -3]]


def test_compute_adj_matrix():
    """
    Test on creating an adjacency matrix from an adjacency array
    """
    adjacencies = [[1, -2], [1, -4], [-1, -3]]
    nb_blocks = 4
    matres = compute_dcj_adjacencies.compute_adjacencymatrix_from_adjacency_array(adjacencies,
                                                                                  nb_blocks)
    mat = SparseMatrix(2 * nb_blocks + 1, 2 * nb_blocks + 1)
    mat[1, 2] = 1
    mat[2, 1] = 1
    mat[1, 4] = 1
    mat[4, 1] = 1
    mat[5, 3] = 1
    mat[3, 5] = 1
    assert matres == mat


def test_writeretained():
    """
    Test write retained dcj adjacencies
    """
    adj_filename_prefix = os.path.join("test", "data", "res_testdcjwriteretained")
    # create empty file for adjacencies in last step (just add the new ones)
    open(adj_filename_prefix + "_1.txt", 'w').close()
    cars = [[], [2, 3], [1], [-4]]
    dcj_NC_Adj = [[2, 3]]
    step = 2
    car_left = [[], [[], [], [], []], [[], [], [], [2]], [[], [], [], []]]
    car_right = [[], [[], [], [], []], [[], [], [3], []], [[], [], [], []]]
    species1_id = [1]
    species2_id = [2]
    species3_id = [3]
    compute_dcj_adjacencies.write_retained_adjs(adj_filename_prefix, cars, dcj_NC_Adj, step,
                                                car_left, car_right, species1_id, species2_id,
                                                species3_id)
    with open(adj_filename_prefix + "_2.txt", 'r') as res:
        lines = res.readlines()
    assert lines == ["1 -4 2 3 2 2 0 1 0\n"]
    os.remove(adj_filename_prefix + "_2.txt")
    os.remove(adj_filename_prefix + "_1.txt")
