# -*- coding: utf-8 -*-:
"""
Unit tests for def_adjacency_functions.py
"""
from procars.utils import util_adjacency_functions
from procars.utils.util_classes import SparseMatrix
import pytest
import os


def test_split_adj_mat():
    """
    Test on splitting an adjacency matrix into a matrix with conflicting adjs and another
    one with non-conflicting adjs
    """
    mat = SparseMatrix(9, 9)
    mat[1, 2] = 1
    mat[2, 1] = 1
    mat[1, 4] = 1
    mat[4, 1] = 1
    mat[3, 5] = 1
    mat[5, 3] = 1
    nb_blocks = 4
    nc, c_mat = util_adjacency_functions.split_adjacencymatrix(mat, nb_blocks)
    expect_nc = SparseMatrix(9, 9)
    expect_nc[3, 5] = 1
    expect_nc[5, 3] = 1
    expect_c = SparseMatrix(9, 9)
    expect_c[1, 2] = 1
    expect_c[2, 1] = 1
    expect_c[1, 4] = 1
    expect_c[4, 1] = 1
    assert nc == expect_nc
    assert c_mat == expect_c


def test_compute_pairs():
    """
    Test on getting all pairs of indexes where sparse_matrix != 0
    """
    mat = SparseMatrix(9, 9)
    mat[1, 2] = 1
    mat[2, 1] = 1
    mat[1, 4] = 1
    mat[4, 1] = 1
    mat[3, 5] = 1
    mat[5, 3] = 1
    pairs = util_adjacency_functions.compute_extremity_pair(mat)
    assert set(pairs) == set([(1, 4), (1, 2), (3, 5)])


def test_compute_adjs():
    """
    Test on computing car adjacencies from their IDs
    """
    extr_pairs = [(1, 4), (1, 2), (3, 5)]
    nb_blocks = 4
    adjs = util_adjacency_functions.compute_adjacencies(extr_pairs, nb_blocks)
    assert adjs == [[1, -4], [1, -2], [3, 1]]


def test_cartobloc_left_pos():
    """
    Test on finding the block at the left end of a given car (positive id)
    """
    cars = [[1, -3, 2, 5], [6, -5], [-7], [9, 11, 8]]
    car = 3
    side = 0
    block = util_adjacency_functions.car_to_block(car, side, cars)
    assert block == 9


def test_cartobloc_right_pos():
    """
    Test on finding the block at the right end of a given car (positive id)
    """
    cars = [[1, -3, 2, 5], [6, -5], [-7], [9, 11, 8]]
    car = 1
    side = 1
    block = util_adjacency_functions.car_to_block(car, side, cars)
    assert block == -5


def test_cartobloc_left_neg():
    """
    Test on finding the block at the left end of a given car (negative id)
    """
    cars = [[1, -3, 2, 5], [6, -5], [-7], [9, 11, 8]]
    car = -3
    side = 0
    block = util_adjacency_functions.car_to_block(car, side, cars)
    assert block == -8


def test_cartobloc_right_neg():
    """
    Test on finding the block at the right end of a given car (negative id)
    """
    cars = [[1, -3, 2, 5], [6, -5], [-7], [9, 11, 8]]
    car = -3
    side = 1
    block = util_adjacency_functions.car_to_block(car, side, cars)
    assert block == -9


def test_conserv_status0():
    """
    Test on finding that a given adjacency is not supported by any group
    """
    # Species tree: ((A, B), ((C, D), (E, F))@);
    # IDs : A=1, B=2, C=3, D=4, E=5, F=6
    # genomes : A: 1 2 3 4
    #           B: 1 -3 -2 4
    #           C: 1 -2 3 4
    #           D: 3 4 1 -2
    #           E: -2 3 4 -1
    #           F: 1 -4 -3 -2
    adj = (1, 4)
    left = [[], [[], [], [1], [2], [3]], [[], [], [-4], [2], [-2]], [[], [], [-3], [-2], [3]],
            [[], [4], [], [], [3]], [[], [], [-3], [-2], [3]], [[], [], [], [2], [3]]]
    right = [[], [[], [2], [3], [4], []], [[], [-3], [3], [-1], []], [[], [-2], [-1], [4], []],
             [[], [-2], [-1], [4], [1]], [[], [-4], [], [4], [-1]], [[], [-4], [3], [4], [-1]]]
    spe1 = [1, 2]
    spe2 = [3, 4]
    spe3 = [5, 6]
    nb, support = util_adjacency_functions.conservation_status(adj, left, right, spe1, spe2, spe3)
    assert nb == 0
    assert support == [0] * 6


def test_conserv_status1():
    """
    Test on finding that a given adjacency is supported by only one group (and one species)
    """
    # same set as test_conserv_status0
    adj = (1, 2)
    left = [[], [[], [], [1], [2], [3]], [[], [], [-4], [2], [-2]], [[], [], [-3], [-2], [3]],
            [[], [4], [], [], [3]], [[], [], [-3], [-2], [3]], [[], [], [], [2], [3]]]
    right = [[], [[], [2], [3], [4], []], [[], [-3], [3], [-1], []], [[], [-2], [-1], [4], []],
             [[], [-2], [-1], [4], [1]], [[], [-4], [], [4], [-1]], [[], [-4], [3], [4], [-1]]]
    spe1 = [1, 2]
    spe2 = [3, 4]
    spe3 = [5, 6]
    nb, support = util_adjacency_functions.conservation_status(adj, left, right, spe1, spe2, spe3)
    assert nb == 1
    assert support == [1, 0, 0, 0, 0, 0]


def test_conserv_status2():
    """
    Test on finding that a given adjacency is supported by 2 different groups (partly conserved)
    with 2 species supporting in one group, one species supporting in the other group.
    """
    # same set as test_conserv_status0
    adj = (2, 3)
    left = [[], [[], [], [1], [2], [3]], [[], [], [-4], [2], [-2]], [[], [], [-3], [-2], [3]],
            [[], [4], [], [], [3]], [[], [], [-3], [-2], [3]], [[], [], [], [2], [3]]]
    right = [[], [[], [2], [3], [4], []], [[], [-3], [3], [-1], []], [[], [-2], [-1], [4], []],
             [[], [-2], [-1], [4], [1]], [[], [-4], [], [4], [-1]], [[], [-4], [3], [4], [-1]]]
    spe1 = [1, 2]
    spe2 = [3, 4]
    spe3 = [5, 6]
    nb, support = util_adjacency_functions.conservation_status(adj, left, right, spe1, spe2, spe3)
    assert nb == 2
    assert support == [1, 1, 0, 0, 0, 1]


def test_conserv_status3():
    """
    Test on finding that a given adjacency is supported by all 3 groups (fully conserved), with
    one or two species supporting in each group
    """
    # same set as test_conserv_status0
    adj = (-4, -3)
    left = [[], [[], [], [1], [2], [3]], [[], [], [-4], [2], [-2]], [[], [], [-3], [-2], [3]],
            [[], [4], [], [], [3]], [[], [], [-3], [-2], [3]], [[], [], [], [2], [3]]]
    right = [[], [[], [2], [3], [4], []], [[], [-3], [3], [-1], []], [[], [-2], [-1], [4], []],
             [[], [-2], [-1], [4], [1]], [[], [-4], [], [4], [-1]], [[], [-4], [3], [4], [-1]]]
    spe1 = [1, 2]
    spe2 = [3, 4]
    spe3 = [5, 6]
    nb, support = util_adjacency_functions.conservation_status(adj, left, right, spe1, spe2, spe3)
    assert nb == 3
    assert support == [1, 0, 1, 1, 1, 1]


def test_findinfo():
    """
    Test on finding information about a given adjacency
    """
    # same set as test_conserv_status0
    adj = (-4, -3)
    left = [[], [[], [], [1], [2], [3]], [[], [], [-4], [2], [-2]], [[], [], [-3], [-2], [3]],
            [[], [4], [], [], [3]], [[], [], [-3], [-2], [3]], [[], [], [], [2], [3]]]
    right = [[], [[], [2], [3], [4], []], [[], [-3], [3], [-1], []], [[], [-2], [-1], [4], []],
             [[], [-2], [-1], [4], [1]], [[], [-4], [], [4], [-1]], [[], [-4], [3], [4], [-1]]]
    spe1 = [1, 2]
    spe2 = [3, 4]
    spe3 = [5, 6]
    labels = util_adjacency_functions.find_adj_info(adj, left, right, spe1, spe2, spe3)
    assert labels == " 1 0 1 1 1 1"
