# -*- coding: utf-8 -*-:
"""
Unit tests for compute_resolved_conflicts.py
"""
from procars.step_modules import compute_resolved_conflicts
import pytest
import os
import shutil
try:
    import cPickle as pickle
except ImportError:
    import pickle


# ########################################################################################
# All fixtures needed for tests
@pytest.fixture(scope="module")
def expect_tree():
    return {1: ['N1', -1, 2, 7], 2: ['N2', 1, 3, 4], 3: ['C_c', 2, -1, -1], 4: ['N4', 2, 5, 6],
            5: ['A_a', 4, -1, -1], 6: ['B', 4, -1, -1], 7: ['N7', 1, 8, 9],
            8: ['D_species', 7, -1, -1], 9: ['E', 7, -1, -1]}


@pytest.fixture(scope="module")
def path():
    return [0, 1, 0, 0, 0, 0, 0, 1, 0, 0]


@pytest.fixture(scope="module")
def leaves():
    return [3, 5, 6, 8, 9]


@pytest.fixture(scope="module")
def internals():
    return [4, 2, 1, 7]


# ########################################################################################
# Unit tests for compute_resolved_conflicts.py
def test_compute_cost(expect_tree, path, leaves, internals):
    """
    Test on calculating the cost of an adjacency according to its presence in the leaves
    """
    labels = {3: [2], 5: [2], 6: [1], 8: [2], 9: [1]}
    nb_labels = 2
    tab = compute_resolved_conflicts.compute_cost(expect_tree, path, leaves,
                                                  internals, labels, nb_labels)
    assert tab == [[9223372036854775807, 0, 0], [9223372036854775807, 2, 1],
                   [9223372036854775807, 2, 1], [9223372036854775807, 9223372036854775807, 0],
                   [9223372036854775807, 1, 1], [9223372036854775807, 9223372036854775807, 0],
                   [9223372036854775807, 0, 9223372036854775807], [9223372036854775807, 3, 2],
                   [9223372036854775807, 9223372036854775807, 0],
                   [9223372036854775807, 0, 9223372036854775807]]


def test_get_adj_costs(leaves, path, internals, expect_tree):
    """
    Test on getting all adjacencies from S_adjacencies file, and finding the cost of each one
    """
    adj_file = 'test/data/res_S_Adjacencies_boreoeutherian_ancestor_2.txt_discarded.txt'
    nb_species = 5
    spe_ids = {'C_c': 0, 'B': 2, 'E': 4, 'D_species': 3, 'A_a': 1}
    ances_id = 7
    res = compute_resolved_conflicts.get_adjacencies_costs(adj_file, nb_species, leaves,
                                                           expect_tree, spe_ids, path,
                                                           internals, ances_id)
    score, ids, infos = res
    assert score == {0: 1, 1: 3, 2: 3, 3: 1, 4: 3, 5: 3}
    assert ids == {0: (134, -142), 1: (134, 396), 2: (142, 134), 3: (611, 612),
                   4: (-612, 589), 5: (-134, 396)}
    assert infos == {0: ['9', '-10', '1', '2', '0', '0', '0', '1', '1'],
                     1: ['9', '27', '1', '2', '0', '1', '1', '0', '0'],
                     2: ['10', '9', '1', '2', '1', '0', '0', '0', '0'],
                     3: ['31', '-34', '1', '2', '0', '0', '0', '1', '1'],
                     4: ['34', '31', '1', '2', '0', '0', '0', '0', '0'],
                     5: ['-9', '27', '1', '2', '0', '0', '0', '0', '0']}


def test_find_conflicts():
    """
    Test on finding couples of adjacencies which are in conflict because they share
    a same end.
    """
    adj_ids = {0: (1, 2), 1: (1, 3), 2: (-1, 2), 3: (-3, 8), 4: (5, 3), 5: (2, 4)}
    conf = compute_resolved_conflicts.find_conflicts_couples(adj_ids)
    assert conf == [(0, 1), (0, 2), (1, 3), (1, 4), (3, 4)]


def test_find_conflicts2():
    """
    Test on finding couples of adjacencies which are in conflict because they share
    a same end.
    """
    adj_ids = {0: (134, -142), 1: (134, 396), 2: (142, 134), 3: (611, 612),
               4: (-612, 589), 5: (-134, 396)}
    conf = compute_resolved_conflicts.find_conflicts_couples(adj_ids)
    assert conf == [(0, 1), (0, 2), (1, 5), (2, 5), (3, 4)]


def test_find_ind_sets():
    """
    Test on finding maximum sized independent sets, when max size is 2 (initial sets)
    """
    nb_adj = 5
    conflicts = [(0, 1), (0, 2), (1, 3), (1, 4), (3, 4)]
    non_conflicts = [(0, 3), (0, 4), (1, 2), (2, 3), (2, 4)]
    sets = compute_resolved_conflicts.find_independant_sets(nb_adj, conflicts)
    assert sets == [list(elem) for elem in non_conflicts]


def test_find_ind_sets2():
    """
    Test on finding maximum sized independent sets when max size > 2
    """
    nb_adj = 6
    conflicts = [(0, 1), (0, 2), (1, 5), (2, 5), (3, 4)]
    sets = compute_resolved_conflicts.find_independant_sets(nb_adj, conflicts)
    assert sets == [[0, 3, 5], [0, 4, 5], [1, 2, 3], [1, 2, 4]]


def test_find_moncost_set():
    """
    Test on finding the set of minimum cost
    """
    independent_sets = [[0, 3, 5], [0, 4, 5], [1, 2, 3], [1, 2, 4]]
    adjacencies_score = [1, 4, 4, 1, 2, 1]
    min_set = compute_resolved_conflicts.find_mincost_set(independent_sets, adjacencies_score)
    assert min_set == [0, 3, 5]


def test_change_adjfiles():
    """
    Test on saving discarded adjacencies into a new file for information, and removing
    them from last file
    """
    adj_prefix = os.path.join("test", "data", "testresolve.changeadjfiles")
    ref_discard = os.path.join("test", "data", "res_S_Adjacencies_discarded_file.txt")
    last_discard = os.path.join("test", "data", "testresolve.changeadjfiles_1.txt_discarded.txt")
    new_discard = os.path.join("test", "data",
                               "testresolve.changeadjfiles_1.txt_discarded_used.txt")
    shutil.copyfile(ref_discard, last_discard)
    step = 1
    compute_resolved_conflicts.change_adj_discarded_files(adj_prefix, step)
    with open(new_discard, "r") as new:
        newlines = new.readlines()
    with open(ref_discard, "r") as ref:
        reflines = ref.readlines()
    assert newlines == reflines
    with open(last_discard, "r") as last:
        assert not last.readlines()
    os.remove(last_discard)
    os.remove(new_discard)


def test_writeretained():
    """
    Test on writting retained adjacencies from conflict into the adjacency file
    """
    step = 2
    adj_prefix = os.path.join("test", "data", "testresolve.writeretained")
    last_adj = os.path.join("test", "data", "testresolve.writeretained_1.txt")
    new_adj = os.path.join("test", "data", "testresolve.writeretained_2.txt")
    ref_adj = os.path.join("test", "data", "res_S_Adjacencies_file.txt")
    adj_ids = {0: (123, 456), 1: (-12, -34), 2: (45, -123), 3: (-65, 48), 4: (12, 3)}
    adj_infos = {0: [2, 3, 1, 1, 0, 0, 1, 1, 1, 0], 1: [-5, 4, 1, 1, 0, 1, 1, 1, 0, 1],
                 2: [6, 5, 0, 1, 1, 1, 0, 0, 0], 3: [-2, 5, 1, 1, 1, 1, 1, 1],
                 4: [6, -8, 1, 1, 0, 0, 0, 1, 1, 1]}
    max_set = [0, 2, 3]
    shutil.copyfile(ref_adj, last_adj)
    compute_resolved_conflicts.write_retained_adjs(adj_prefix, adj_infos, max_set, adj_ids, step)
    with open(last_adj, 'r') as last:
        lastlines = last.readlines()
    with open(ref_adj, 'r') as ref:
        reflines = ref.readlines()
    assert lastlines == reflines
    with open(new_adj, "r") as new:
        newlines = new.readlines()
    expect = ["123 456 2 3 1 1 0 0 1 1 1 0\n", "45 -123 6 5 0 1 1 1 0 0 0\n",
              "-65 48 -2 5 1 1 1 1 1 1\n"] + reflines
    assert newlines == expect
    os.remove(last_adj)
    os.remove(new_adj)
