# -*- coding: utf-8 -*-:
"""
Unit tests for utils_classes.py
"""

from procars.utils.util_classes import SparseMatrix
import pytest
import os


def test_getelem():
    """
    Test the get_item function of sparse_matrix class
    """
    lines = 4
    cols = 5
    mat = SparseMatrix(lines, cols)
    mat[1, 2] = 10
    mat[3, 4] = 5
    # get existing elements
    assert mat[1, 2] == 10
    assert mat[3, 4] == 5
    # get non-existing element
    assert mat[1, 3] == 0
    # get whole line with one existing element
    assert mat[1] == [10]
    assert mat[3] == [5]
    # get whole line with no element
    assert mat[2] == []


def test_get_out_elem():
    """
    Test on trying to get an outbound element of sparse_matrix
    """
    lines = 4
    cols = 5
    mat = SparseMatrix(lines, cols)
    # element with line out of bounds
    with pytest.raises(IndexError) as excinfo:
        mat[5, 0]
    assert excinfo.value.message == "row index out of bounds"
    # element with column out of bounds
    with pytest.raises(IndexError) as excinfo:
        mat[2, 10]
    assert excinfo.value.message == "column index out of bounds"
    # whole line out of bounds
    with pytest.raises(IndexError) as excinfo:
        mat[5]
    assert excinfo.value.message == "row index out of bounds"


def test_setelem():
    """
    Test set_item function of sparse_matrix class
    """
    lines = 4
    cols = 5
    mat = SparseMatrix(lines, cols)
    assert mat[1, 2] == 0
    mat[1, 2] = 5
    assert mat[1, 2] == 5


def test_set_out_item():
    """
    Test on trying to set an element out of the sparse_matrix bounds
    """
    lines = 4
    cols = 5
    mat = SparseMatrix(lines, cols)
    with pytest.raises(IndexError) as excinfo:
        mat[5, 0] = 2
    assert excinfo.value.message == "row index out of bounds"
    with pytest.raises(IndexError) as excinfo:
        mat[2, 10] = 6
    assert excinfo.value.message == "column index out of bounds"
    with pytest.raises(IndexError) as excinfo:
        mat[5] = [5, 6, 2]
    assert excinfo.value.message == "give (line, col) index"


def test_print_sparsematrix(capsys):
    """
    Test on printing a sparse matrix
    """
    lines = 4
    cols = 5
    mat = SparseMatrix(lines, cols)
    mat[1, 2] = 10
    mat[3, 4] = 5
    print mat
    out, _ = capsys.readouterr()
    assert "(1, 2): 10\n(3, 4): 5" in out


def test_sumline():
    """
    Test on a sum of all elements in a given line of sparse_matrix
    """
    lines = 4
    cols = 5
    mat = SparseMatrix(lines, cols)
    mat[1, 2] = 10
    mat[3, 4] = 5
    assert sum(mat[1]) == 10
    assert sum(mat[2]) == 0


def test_equals():
    """
    Tests that two SparseMatrices are equal
    """
    lines = 4
    cols = 5
    mat1 = SparseMatrix(lines, cols)
    mat1[1, 2] = 10
    mat1[3, 4] = 5
    mat2 = SparseMatrix(lines, cols)
    mat2[1, 2] = 10
    mat2[3, 4] = 5
    assert mat1 == mat2


def test_not_equal():
    """
    Tests that two SparseMatrices are not the same
    """
    lines = 4
    cols = 5
    mat1 = SparseMatrix(lines, cols)
    mat1[1, 2] = 10
    mat1[3, 4] = 5
    mat2 = SparseMatrix(lines, cols)
    mat2[1, 2] = 10
    mat2[3, 4] = 5
    mat2[1, 3] = 1
    assert not mat1 == mat2


def test_getpairs():
    """
    Test on getting all pairs (x, y) of indexes where sparse_matrix != 0
    """
    lines = 4
    cols = 5
    mat1 = SparseMatrix(lines, cols)
    mat1[1, 2] = 10
    mat1[3, 4] = 5
    mat1[2, 3] = 4
    assert set(mat1.pairs()) == set([(1, 2), (3, 4), (2, 3)])
